#
# VT50-2000-2016-84-28-4
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC. All rights reserved.
# 
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF VERITAS
# TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE
# PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
# 
# The Licensed Software  and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software  -  Restricted Rights"
# and DFARS 227.7202,  Rights  in  "Commercial Computer Software or  Commercial
# Computer Software Documentation" as applicable, and any successor regulations,
# whether  delivered  by  Veritas  as  on premises or hosted services.  Any use,
# modification, reproduction release, performance, display or disclosure of the
# Licensed Software and Documentation by the U.S. Government shall be solely in
# accordance with the terms of this Agreement. $
#
package CPIC;
use strict;

sub uninstall {
    my $cpic=shift;
    my ($phase,$rtn,$prod,$rel);

    $rel=$cpic->rel;
    $prod=$rel->get_operate_product($cpic->prod,'uninstall');

    $rtn=$cpic->preuninstall($prod);
    if ($rtn) {
        $phase=$cpic->setup_uninstall_phase($prod);
        $phase->execute_tasks();
        $rtn=$cpic->postuninstall($prod);
    }
   
    return $rtn;
}

# Return 1, continue uninstall
# Return 0, skip whole uninstall phase.
sub preuninstall {
    my ($cpic,$prod)=@_;

    # If -makeresponsefile but no prod specified, skip whole uninstall phase.
    return 0 if (Cfg::opt('makeresponsefile') && !($prod));

    $cpic->{falsefailures}=0;

    $cpic->{uninstallbundles}=$cpic->verify_list('uninstallbundles');
    $cpic->{uninstallpkgs}=$cpic->verify_list('uninstallpkgs');

    my $steps=scalar(@{$cpic->{uninstallbundles}}) + scalar(@{$cpic->{uninstallpkgs}});

    # If no uninstall tasks, skip whole uninstall phase.
    return 0 if (!$steps);

    $prod->preremove_tasks if $prod->can('preremove_tasks');

    # Show uninstall UI form.
    $cpic->show_ui_form('uninstalling');
    return 1;
}

sub setup_uninstall_phase {
    my ($cpic,$prod)=@_;
    my ($phase,$phase_name,$phase_desc,$task,$task_name,%task_params,$pkg,$rel,$pdfrs,$sub,$id);
    $rel=$cpic->rel;

    $phase_name=$cpic->{uninstall_phase_name} || 'CPIC_uninstall';
    $phase_desc=$cpic->{uninstall_phase_desc} || Msg::new("Uninstalling $prod->{abbr}");

    $phase=Phase->new($phase_name);
    $phase->set_description($phase_desc);
    $phase->initialize_tasks();

    # Perform preremove tasks
    $id=10;
    $task_name = "product_preremove";
    %task_params=(
        'sequence_id' => $id,
        'description' => Msg::new("Performing $prod->{abbr} preremove tasks"),
        'description_sys' => Msg::new("Performing $prod->{abbr} preremove tasks on #{SYS}"),
        'if_per_system' => 1,
        'handler' => \&CPIC::prod_preremove_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::prod_preremove_errors,
        'post_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    # Uninstall bundles
    $id+=100;
    for my $bundle (@{$cpic->{uninstallbundles}}) {
        $task_name = "uninstall_bundle_$bundle";
        %task_params=(
            'sequence_id' => $id,
            'description' => Msg::new("Uninstalling $bundle bundle"),
            'description_sys' => Msg::new("Uninstalling $bundle bundle on #{SYS}"),
            'if_per_system' => 1,
            'handler' => \&CPIC::bundle_uninstall_sys,
            'handler_object_arg' => $cpic,
            'handler_args' => [ $bundle ],
        );
        $task=Task->new($task_name, %task_params);
        $phase->add_task($task);

        # Skip the task on some systems for serial mode
        for my $sys (@{$cpic->{systems}}) {
            if (!EDRu::inarr($bundle, @{$sys->{uninstallbundles}})) {
                $task->set_skip_sys($sys,1);
            }
        }
        $id++;
    }

    # Uninstall packages
    $id+=100;
    if ($rel->{batchuninstall}) {
        if ($cpic->{uninstallpkgs}) {
            $pdfrs=Msg::get('pdfrs');
            $task_name = "uninstall_pkgs";
            %task_params=(
                'sequence_id' => $id,
                'description' => Msg::new("Uninstalling $prod->{abbr} $pdfrs"),
                'description_sys' => Msg::new("Uninstalling $prod->{abbr} $pdfrs on #{SYS}"),
                'if_per_system' => 1,
                'handler' => \&CPIC::pkgs_uninstall_sys,
                'handler_object_arg' => $cpic,
            );
            $task=Task->new($task_name, %task_params);
            $phase->add_task($task);
        }
    } else {
        for my $pkgi (@{$cpic->{uninstallpkgs}}) {
            $pkg=$cpic->pkg($pkgi);
            $task_name = "uninstall_pkg_$pkgi";
            %task_params=(
                'sequence_id' => $id,
                'description' => Msg::new("Uninstalling $pkg->{pkg}"),
                'description_sys' => Msg::new("Uninstalling $pkg->{pkg} on #{SYS}"),
                'if_per_system' => 1,
                'handler' => \&CPIC::pkg_uninstall_sys,
                'handler_object_arg' => $cpic,
                'handler_args' => [ $pkgi ],
                'rc_handler' => \&CPIC::pkg_uninstall_rc_check_sys,
                'rc_handler_object_arg' => $cpic,
                'rc_handler_args' => [ $pkgi ],
            );
            $task=Task->new($task_name, %task_params);
            $phase->add_task($task);

            # Skip the task on some systems for serial mode
            for my $sys (@{$cpic->{systems}}) {
                if (!EDRu::inarr($pkgi, @{$sys->{uninstallpkgs}})) {
                    $task->set_skip_sys($sys,1);
                }
            }
            $id++;
        }
    }

    # Cleanup CPI scripts
    $id+=100;
    if (Cfg::opt('uninstall') && ((@{$cpic->{uninstallbundles}}) || (@{$cpic->{uninstallpkgs}}))) {
        # only remove if there are no failures, but we still have
        # to go into this block to not screw up the step count

        if ($cpic->rel->{'cpipkg'}) {
            $sub=\&CPIC::readd_install_scripts_sys;
        } else {
            $sub=\&CPIC::cleanup_cpi_sys;
        }
        $task_name = "uninstall_cpi_scripts";
        %task_params=(
            'sequence_id' => $id,
            'description' => Msg::new("Removing install scripts"),
            'description_sys' => Msg::new("Removing install scripts on #{SYS}"),
            'if_per_system' => 1,
            'handler' => $sub,
            'handler_object_arg' => $cpic,
        );
        $task=Task->new($task_name, %task_params);
        $phase->add_task($task);
    }

    # Perform postremove tasks
    $id+=100;
    $task_name = "product_postremove";
    %task_params=(
        'sequence_id' => $id,
        'description' => Msg::new("Performing $prod->{abbr} postremove tasks"),
        'description_sys' => Msg::new("Performing $prod->{abbr} postremove tasks on #{SYS}"),
        'if_per_system' => 1,
        'handler' => \&CPIC::prod_postremove_sys,
        'handler_object_arg' => $cpic,
        'post_handler' => \&CPIC::prod_postremove_errors,
        'post_handler_object_arg' => $cpic,
    );
    $task=Task->new($task_name, %task_params);
    $phase->add_task($task);

    return $phase;
}

sub postuninstall {
    my ($cpic,$prod)=@_;
    my ($msg,$failmsg);

    $msg=Msg::new("$prod->{name} Uninstall completed successfully");
    $failmsg=Msg::new("$prod->{name} Uninstall did not complete successfully");
    $msg->display_completion($cpic->failures('uninstallfailmsg'), $msg, $failmsg);

    $cpic->update_summary_uninstall();
    $cpic->ask_uninstall_continue();

    return 1;
}

sub pkgs_uninstall_sys {
    my ($cpic,$sys)= @_;
    my ($uof,$pkg,$pkgs,$pkgis);
    # scalars must be passed to threaded call
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return 1 if (Cfg::opt('makeresponsefile'));

    $pkgs='';
    $pkgis='';
    for my $pkgi (@{$cpic->{uninstallpkgs}}) {
        next if (!EDRu::inarr($pkgi, @{$sys->{uninstallpkgs}}));
        $pkg=$sys->pkg($pkgi);
        $pkg->preremove_sys($sys);
        $pkgs.=" $pkg->{pkg}";
        $pkgis.=" $pkgi";
    }
    return 1 unless ($pkgs);

    # uninstall packages batchly
    Msg::log("BEGIN-pkgs_uninstall_sys($sys->{sys},'$pkgs')");

    $sys->uninstall_pkgs($pkgs);

    $uof=EDRu::outputfile('uninstall', $sys->{sys}, 'pkgs');
    $sys->copy_to_sys($cpic->localsys,$uof) unless ($sys->{islocal});
    for my $pkgi (@{$cpic->{uninstallpkgs}}) {
        next if (!EDRu::inarr($pkgi, @{$sys->{uninstallpkgs}}));
        $pkg=$sys->pkg($pkgi);
        $pkg->postremove_sys($sys);
    }
    Msg::log("END-pkgs_uninstall_sys($sys->{sys},'$pkgs')");

    return 1 if ($cpic->pkgs_uninstall_verify_sys($sys,$pkgis));
    return 0;
}

sub pkgs_uninstall_verify_sys {
    my ($cpic,$sys,$pkgis)= @_;
    my ($fails,$pkg,$pdfr,$msg);
    $pkgis=~s/^\s*//;
    $pkgis=~s/\s*$//;
    $fails=0;
    $pdfr=Msg::get('pdfr');
    for my $pkgi (split(/\s+/,$pkgis)) {
        $pkg=$sys->pkg($pkgi);
        if ($pkg->uninstall_success_sys($sys)) {
            $msg=Msg::new("$pkg->{pkg} $pdfr uninstalled successfully on $sys->{sys}");
            $msg->log;
            $sys->set_value('pkguninstallpass', 'push', $pkg->{pkg});
            EDRu::appendfile($pkg->{pkg}, EDR::tmpdir()."/pkg_uninstall_pass.$sys->{sys}")
                if (Cfg::opt('bait'));
        } else {
            $msg=Msg::new("$pkg->{pkg} $pdfr failed to uninstall on $sys->{sys}");
            $msg->log;
            $sys->set_value('uninstallfailmsg', 'push', $msg->{msg});
            $sys->set_value('pkguninstallfail', 'push', $pkg->{pkg});
            EDRu::appendfile($pkg->{pkg}, EDR::tmpdir()."/pkg_uninstall_fail.$sys->{sys}")
                if (Cfg::opt('bait'));
            $fails++;
        }
    }
    return 0 if ($fails);
    return 1;
}

sub pkg_uninstall_sys {
    my ($cpic,$sys,$pkg)= @_;
    # scalars must be passed to threaded call
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    return 1 if (Cfg::opt('makeresponsefile'));
    # pkg uninstall is conditional
    return if (!EDRu::inarr($pkg->{pkgi}, @{$sys->{uninstallpkgs}}));
    Msg::log("BEGIN-pkg_uninstall_sys($pkg->{pkg},$sys->{sys})");

    $pkg->preremove_sys($sys);
    $pkg->uninstall_sys($sys);
    my $uof=EDRu::outputfile('uninstall', $sys->{sys}, $pkg->{pkg});
    $sys->copy_to_sys($cpic->localsys,$uof) unless ($sys->{islocal});
    $pkg->postremove_sys($sys);
    Msg::log("END-pkg_uninstall_sys($pkg->{pkg},$sys->{sys})");
    # Adding temp failures to test for web script

    return 0 if ($cpic->false_failure);
    return 1 if ($pkg->uninstall_success_sys($sys));
    return 0;
}

sub pkg_uninstall_failed_sys {
    my($sys,$pkg)= @_;
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    push(@{$sys->{pkguninstallfail}}, $pkg->{pkg});
    my $pdfr=Msg::get('pdfr');
    my $io=EDRu::readfile(EDRu::outputfile('uninstall', $sys->{sys}, $pkg->{pkg}));
    my $msg=Msg::new("$pkg->{pkg} $pdfr failed to uninstall on $sys->{sys}");
    $msg->log;
    Msg::log("$io");
    push(@{$sys->{uninstallfailmsg}}, $msg);
    EDRu::appendfile($pkg->{pkg},
        EDR::tmpdir()."/pkg_uninstall_fail.$sys->{sys}") if (Cfg::opt('bait'));
    return '';
}

sub pkg_uninstall_passed_sys {
    my($sys,$pkg)= @_;
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    # pkg uninstall is conditional
    return '' if (!EDRu::inarr($pkg->{pkgi}, @{$sys->{uninstallpkgs}}));
    push(@{$sys->{pkguninstallpass}}, $pkg->{pkg});
    my $pdfr=Msg::get('pdfr');
    Msg::log("$pkg->{pkg} $pdfr uninstalled successfully on $sys->{sys}");
    EDRu::appendfile($pkg->{pkg},
        EDR::tmpdir()."/pkg_uninstall_pass.$sys->{sys}") if (Cfg::opt('bait'));
    return '';
}

sub pkg_uninstall_rc_check_sys {
    my ($cpic,$sys,$pkg,$rc)= @_;
    $pkg=$sys->pkg($pkg) unless (ref($pkg));
    if ($rc) {
        pkg_uninstall_passed_sys($sys, $pkg);
    } elsif (defined $rc) {
        pkg_uninstall_failed_sys($sys, $pkg);
    }
    return $rc;
}

sub bundle_uninstall_sys {
    my ($cpic,$sys,$bundlei)= @_;
    # scalars must be passed to threaded call
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return 1 if (Cfg::opt('makeresponsefile'));
    # pkg uninstall is conditional
    return 1 if (!EDRu::inarr($bundlei, @{$sys->{uninstallbundles}}));
    Msg::log("BEGIN-bundle_uninstall_sys($bundlei,$sys->{sys})");
    $sys->padv->bundle_uninstall_sys($sys,$bundlei)
        if ($sys->padv->can('bundle_uninstall_sys'));
    Msg::log("END-bundle_uninstall_sys($bundlei,$sys->{sys})");
    return 1;
}

sub prod_preremove {return 1;}

sub prod_preremove_sys {
    my ($cpic,$sys) = @_;
    my ($tmpsub,$prod);
    return 1 if (Cfg::opt('makeresponsefile'));
    # scalars must be passed to threaded call
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $sys->rel->preremove_sys($sys);
    $prod=$sys->prod;
    $tmpsub = upgrade_sub('preremove_sys');
    if ($prod->can($tmpsub)) {
        $prod->$tmpsub($sys);
    }
    return 1;
}

sub prod_preremove_errors {
    my ($cpic) = @_;
    my $prod=$cpic->prod;
    $prod->preremove_errors($cpic) if ($cpic->sys_error);
    EDR::reset_errors_warnings();
    return 1;
}

sub prod_postremove_sys {
    my ($cpic,$sys) = @_;
    my ($tmpsub,$prod);
    return 1 if (Cfg::opt('makeresponsefile'));
    # scalars must be passed to threaded call
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $prod=$sys->prod;
    $tmpsub = upgrade_sub('postremove_sys');
    if ($prod->can($tmpsub)) {
        $prod->$tmpsub($sys);
    }
    $sys->rel->postremove_sys($sys);
    return 1;
}

sub prod_postremove_errors {
    my ($cpic) = @_;
    my $prod=$cpic->prod;
    $prod->postremove_errors($cpic) if ($cpic->sys_error);
    EDR::reset_errors_warnings();
    return 1;
}

sub cleanup_cpi_sys {
    my ($cpic,$sys) = @_;
    my ($rootpath,$rel,$use,$edr,$subprod,$prodname,$prod);
    return 1 if (Cfg::opt('makeresponsefile'));
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $cpic=Obj::cpic() unless (ref($cpic) eq 'CPIC');
    $rel=$sys->rel;
    $prod=$sys->prod;
    $edr=Obj::edr();
    $rootpath=Cfg::opt('rootpath')||'';

    # remove scripts
    for my $subprod ($cpic->{prod}, @{$prod->{subprods}}) {
        $prod=$sys->prod($subprod);
        $prodname=lc($prod->{installscript_name} || $prod->{prod});
        $prodname=~s/ /_/mg;
        $prodname="_$prodname" if (length($prodname)>8);
        $sys->rm("$rootpath$cpic->{installpath}/install$prodname");
        $sys->rm("$rootpath$cpic->{installpath}/uninstall$prodname");
    }

    # remove perl libs if there are no other versions using it
    # do not remove perl libs if it's running under localized locale
    $use=$sys->grepfile("$cpic->{release}","$rootpath$cpic->{installpath}/*");
    $sys->rm("$rootpath$cpic->{perlpath}/lib/site_perl/$cpic->{release}")
        if ((!$use) && ($edr->{envlang} eq 'C'));
    return 1;
}

# delete all supported install/uninstall* scripts
# and then add them back for still installed products.
sub readd_install_scripts_sys {
    my ($cpic,$sys) = @_;
    my ($script,$rootpath);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $rootpath=Cfg::opt('rootpath');
    $script=$rootpath.$cpic->{installpath}.'/bin/'.$cpic->{release}.'/add_install_scripts';
    # if $script doesn't exist, all scripts should have been removed when CPI package was uninstalled.
    if ($sys->exists($script)) {
        $sys->cmd("CALLED_BY=CPI $script delete");
        $sys->cmd("CALLED_BY=CPI DONOTUPDATEVH=1 $script");
    }
    return 1;
}

sub update_summary_uninstall {
    my $cpic=shift;
    my ($msg,$pdfrs);

    $pdfrs=Msg::get('pdfrs');
    for my $sys (@{$cpic->{systems}}) {
        if ($#{$sys->{pkguninstallpass}}>=0) {
            $msg=Msg::new("The following $pdfrs were uninstalled on $sys->{sys}:");
            $cpic->update_summary($msg, @{$sys->{pkguninstallpass}});
        }
        if ($#{$sys->{pkguninstallfail}}>=0) {
            $msg=Msg::new("The following $pdfrs failed to uninstall on $sys->{sys}:");
            $cpic->update_summary($msg, @{$sys->{pkguninstallfail}});
        }
    }
    return '';
}

sub ask_uninstall_continue {
    my $cpic = shift;
    my ($msg, $msg_advice, $pdfrs, $ayn, $pkguninstallfail_flag);
    my $edr = Obj::edr();
    for my $sys (@{$cpic->{systems}}) {
        if ($#{$sys->{pkguninstallfail}}>=0) {
            $pkguninstallfail_flag = 1;
        }
    }
    if ($pkguninstallfail_flag == 1){
        $pdfrs=Msg::get('pdfrs');
        $msg = Msg::new("$pdfrs failed to uninstall. Do you want to exit $cpic->{script}?");
        $ayn = $msg->ayny;
        Msg::n();
        if ($ayn eq 'Y') {
            $msg_advice = Msg::new("Refer to the installer logs for details and correct the settings. Continue with the installer again, or try to uninstall the $pdfrs manually");
            $msg_advice->print;
            Msg::n();
            $cpic->edr_completion();
        }
    }
    return '';
}

1;

