#
# VT50-2000-2016-84-28-4
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC. All rights reserved.
# 
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF VERITAS
# TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE
# PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
# 
# The Licensed Software  and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software  -  Restricted Rights"
# and DFARS 227.7202,  Rights  in  "Commercial Computer Software or  Commercial
# Computer Software Documentation" as applicable, and any successor regulations,
# whether  delivered  by  Veritas  as  on premises or hosted services.  Any use,
# modification, reproduction release, performance, display or disclosure of the
# Licensed Software and Documentation by the U.S. Government shall be solely in
# accordance with the terms of this Agreement. $
#
package Rel::UXRT731::Common;
use strict;

# main sub for product licensing
# Cfg::opt('nolic')       : skip licensing
# $cfg->{keys}{keyless}   : keys array for keyless keys
# $cfg->{keys}{license}   : keys array for user-defined keys

sub licensing {
    my ($rel) = @_;
    my ($cpic,$msg,$cfg,$licensed,$rt,$prod);
    my ($vlic_uninstalled, $unlicensed);
    my ($keyless_exists,$license_exists);
    
    $cpic=Obj::cpic();
    $cfg=Obj::cfg();
    $prod = $cpic->prod;
    
    # Make change for E2852406
    if (Cfg::opt('nolic')) {
        $msg = Msg::new("Configuration, startup, or installation of license-based features are skipped when using the -nolic option.");
        $msg->bold();
        $cpic->completion();
    }
    # Do NOT entry licensing process in Live Upgrade or $prod->{nolic} set
    return '' if ($prod->{nolic});
    # Do NOT use default value when licensing
    Cfg::unset_opt('silent');

    # Set general result for $rel->{license_status}
    $rel->license_status();
    # if at least one system doesn't have VRTSvlic installed, report the issue.
    if (scalar(@{$rel->{license_status}{vlic_uninstalled}}) > 0) {
        $vlic_uninstalled=join(',',@{$rel->{license_status}{vlic_uninstalled}});
        $msg=Msg::new("The following systems don't have VRTSvlic installed: $vlic_uninstalled. CPI licensing part will be skipped.");
        $msg->bold();
        return '';
    }
    if (scalar(@{$rel->{license_status}{unlicensed}}) > 0) {
        $unlicensed=join(',',@{$rel->{license_status}{unlicensed}});
        Msg::log("The following systems don't have license key registered for $prod->{prod}: $unlicensed.");
    }
    if (Cfg::opt("responsefile")) {
        # with responsefile, check keyless/permanent keys, register them and then return.
        if ($cfg->{keys}{keyless}) {
            # keyless, register all the keys from $cfg->{keys};
            $rel->responsefile_keyless_key_registry($prod);
        }
        if ($cfg->{keys}{license}) {
            # permanent key, register all the keys from $cfg->{keys}
            $rel->responsefile_permanent_key_registry($prod);
        }
        return 1;
    }
    my ($result,$key);     
    if (Cfg::opt(qw/install license/)) {
        $rel->license_menu();
    } elsif (Cfg::opt(qw/configure start/)) {
        #During configure or start process, prompt customer if the prod wasn't licensed and the previous key is keyless.
        if (scalar(@{$rel->{license_status}{unlicensed}}) > 0) {
            if ($rel->previous_keyless_key_exists()) {
                $msg=Msg::new("Found vxkeyless key(s) for previous release(s). They will be automatically updated to the current version.");
                $msg->print();
                $rel->keyless_key_registry($prod);
            }
        }
    } elsif (Cfg::opt(qw/upgrade patchupgrade hotfixupgrade/)) {
        $keyless_exists=$rel->keyless_key_exists();
        $license_exists=$rel->permanent_key_exists();
        # We cannot determine if both keyless and permanent key exist for now.
        if (!$keyless_exists && !$license_exists && !$rel->is_single_node_cps_key()) {
            $rel->license_menu();
        } elsif ($keyless_exists) {
            if ($rel->previous_keyless_key_exists()) {
                $msg=Msg::new("Found vxkeyless key(s) for previous release(s). They will be automatically updated to the current version.");
                $msg->print();
                $rel->keyless_key_registry($prod);
            }
        } elsif ($license_exists) {
            if ($rel->previous_permanent_key_exists()) {
                $msg=Msg::new("Found user defined key(s) for previous release(s). Would you like to update the key?");
                while (1) {
                    $result=$msg->ayny();
                    if ($result eq 'Y') {
                        $result=$rel->permanent_key_registry($prod);
                        if (EDR::getmsgkey($result,'back')) {
                            Msg::n();
                            next;
                        }
                    } else {
                        Msg::n();
                        $msg=Msg::new("Full $prod->{abbr} functionality will not be enabled until a new $rel->{license_ver} key has been registered\n");
                        $msg->warning();
                    }
                    last;
                }
            }
        }
    }
    # Handle the following scenario:
    # Enterprise installed with Availability license, but user actually tries to configure SF.
    # Give a warning message and ask if user would like to continue
    # CPS only use part of VCS funcionality and do not have VCS key, skip license checking
    if (!Cfg::opt("responsefile") && $prod->{activecomponent}) {
        my $activecomponent=$prod->prod(${$prod->{activecomponent}}[0]);
        # only test the first system if activecomponent is licensed, save time.
        while (!$activecomponent->licensed_sys(${$cpic->{systems}}[0])) {
            # the activecomponent is not licensed. Ask user to retry.
            $msg=Msg::new("$activecomponent->{prod} component cannot be started because it is not covered by the license installed on the system. Would you like to enter the license menu?");
            my $ayn=$msg->ayny;
            if ($ayn eq 'Y') {
                $rel->license_menu();
            } else {
                Msg::n();
                last;
            }
        }
    }
    return 1;
}

sub is_single_node_cps_key_sys {
    my ($rel,$sys) = @_;
    my $pkg=$sys->pkg('VRTScps731');
    if ($pkg && $sys->{keys} && scalar(keys%{$sys->{keys}})>0) {
        for my $one_key (keys %{$sys->{keys}}) {
            return 1 if($pkg->{cpskey} eq $sys->{keys}{$one_key}{'License Key'});
        }
    }
    return 0;
}

sub is_single_node_cps_key {
    my $rel = shift;
    my $cpic=Obj::cpic();
    if ($cpic->{systems} && scalar @{$cpic->{systems}} == 1) {
        my $sys = $cpic->{systems}[0];
        return 1 if ($rel->is_single_node_cps_key_sys($sys));
    }
    return 0;
}

# return a license hash including license type, license product name and license key.
sub license_menu {
    my ($rel)=@_;
    my ($msg,$cfg,$cpic,$prod,$intromsg);
    my ($type,$menumsg,$help,$license);

    $cfg=Obj::cfg();
    $cpic=Obj::cpic();
    $prod=$cpic->prod;

    # Select product license type : vxkeyless or enter a license key
    $msg=Msg::new("Enter a valid license key");
    push(@{$menumsg},$msg->{msg});
    $msg=Msg::new("Enable keyless licensing and complete system licensing later");
    push(@{$menumsg},$msg->{msg});
    $intromsg=Msg::new("To comply with the terms of our End User License Agreement, you have 60 days to either:\n\n * Enter a valid license key matching the functionality in use on the systems\n * Enable keyless licensing and manage the systems with a Management Server. For more details visit http://www.veritas.com/community/blogs/introducing-keyless-feature-enablement-storage-foundation-ha-51. The product is fully functional during these 60 days.\n");
    $msg=Msg::new("How would you like to license the systems?");
    
    while (1) {
        Msg::title();
        $intromsg->print;
        $type=$msg->menu($menumsg, '2', $help, 0);
        if ($type eq '1') {
            $license = $rel->permanent_key_registry($prod);
        } elsif ($type eq '2') {
            $license = $rel->keyless_key_registry($prod);
        }
        # if user choose back, print the menu again
        last unless (EDR::getmsgkey($license,'back')); 
    }
    return $license;
}

# Specific function handling:
# 1. ask for keyless key (designed for Enterprise product)
# 2. register retrieved key on $cpic->{systems} one by one
sub keyless_key_registry {
    my ($rel,$prod)=@_;
    my ($key,$cfg);
    $cfg=Obj::cfg();
    $key=$rel->ask_for_keyless_key($prod);
    return $key if (EDR::getmsgkey($key,'back'));
    $rel->register_keyless_key($prod,$key);
    return 1;
}

sub responsefile_keyless_key_registry {
    my ($rel,$prod)=@_;
    my $cfg=Obj::cfg();
    if ($cfg->{keys}{keyless} && @{$cfg->{keys}{keyless}}) {
        for my $key (@{$cfg->{keys}{keyless}}) {
            $rel->register_keyless_key($prod,$key);
        }
    } else {
        Msg::log("\$CFG{keys}{keyless} not set, skip registering keyless key on the system.");
    }
    return 1;
}

sub ask_for_keyless_key {
    my ($rel,$prod)=@_;
    my ($key,@prods,@names,$menumsg,$msg,$help,$choice);
    if ($prod->{prod} =~ /Foundation|Storage|Availability/mxi) {
        $key=$prod->{lic_keyless_name};
    } else {
        # give a menu
        for my $prodname (qw/FOUNDATION731 AVAILABILITY731 STORAGE731 ENTERPRISE731/) {
            push(@prods,$rel->prod($prodname));
        }
        @names=map({$_->{name}} @prods);
        $menumsg=\@names;
        Msg::n();
        $msg=Msg::new("Which product would you like to register?");
        if (Cfg::opt(qw/upgrade patchupgrade hotfixupgrade/)) {
            $choice=4;
        } else {
            # we need to prompt this message within non-upgrade scenario
            # in other scenarios we directly register it with the $product key
            $choice=$msg->menu($menumsg, '4', $help, 1);
        }
        return $choice if (EDR::getmsgkey($choice,'back'));
        #todo get the product name and then get the lic_names
        $key=$prods[$choice-1]{lic_keyless_name};
    }
    return $key;
}

sub register_keyless_key {
    my ($rel,$prod,$key)=@_;
    my ($msg,$cpic,$cfg,@keylist);
    my ($sys0,$data,$datafile,%keylesshash);
    my ($exit,$result);
    $cpic=Obj::cpic();
    $cfg=Obj::cfg();
    $sys0=${$cpic->{systems}}[0];
    $datafile=$rel->get_lic_data_file_path("/etc/vx/licenses/dat/licenses.dat");
    # Retrieve the actual key from licenses.dat
    if ($sys0->exists($datafile)) {
        $data=$sys0->readfile($datafile);
        for my $line (split(/\n/,$data)) {
            # match the key name and key value
            if ($line =~ /^(\w+).*\s+(\S+)$/) {
                $keylesshash{$1}=$2;
            }
        }
    }
    unless (%keylesshash) {
        $msg=Msg::new("Failed to retrieve keyless key information from $datafile on $sys0->{sys}, aborting keyless key registeration\n");
        $msg->print();
        return 1;
    }
    $msg=Msg::new("Registering keyless key $key on $prod->{name}");
    $msg->print();
    for my $sys (@{$cpic->{systems}}) {
        @keylist=split(/,/, $sys->{vxkeyless});
        if (EDRu::inarr($key, @keylist)) {
            # skip register this key
            $msg=Msg::new("$key is already registered on $sys->{sys}, skip registering this key on $sys->{sys}");
            $msg->print();
            next;
        }
        ($exit,$result)=$rel->register_key_sys($sys,$keylesshash{$key});
        if ($exit) {
            $msg=Msg::new("Failed to register $key keyless key on $sys->{sys}:\n$result");
        } else {
            $msg=Msg::new("Successfully registered $key keyless key on $sys->{sys}");
        }
        $msg->print();
    }
    Msg::n();
    $cfg->set_value("keys,keyless","push",$key) unless (EDRu::inarr($key,@{$cfg->{keys}{keyless}}));
    return 1;
}

sub register_keyless_key_sys {
    my ($rel,$sys,$key)=@_;
    my ($command);
    # register $prod->{lic_keyless_name} key
    $command = $rel->get_lic_binary_path("vxkeyless set -q $key");
    $sys->cmd("$command 2>/dev/null");
    # each time when a new key is registered, reset $sys->{donotquerykeys}
    undef ($sys->{donotquerykeys});
    return 1;
}

sub responsefile_permanent_key_registry {
    my ($rel,$prod)=@_;
    my ($cfg,$prodname);
    $cfg=Obj::cfg();
    if ($cfg->{keys}{license} && @{$cfg->{keys}{license}}) {
        for my $key (@{$cfg->{keys}{license}}) {
            $prodname=$rel->validate_permanent_key($key,$prod);
            if ($prodname) {
                $rel->register_permanent_key($prod,$prodname,$key);
            } else {
                Msg::log("$key is not a valid key defined in the responsefile, skip registering this key");
            }
        }
    } else {
        Msg::log("\$CFG{keys}{license} not set, skip registering user-defined key.");
    }
    return 1;
}
            
# for enterprise, accept multiple permanent keys
# for other products, only accept the permanent key of the product
sub permanent_key_registry {
    my ($rel,$prod)=@_;
    my ($keyprodname,$key,$msg,$ayn);
    ($keyprodname,$key)=$rel->ask_for_permanent_key($prod);
    return $keyprodname if (EDR::getmsgkey($keyprodname,'back'));
    $rel->register_permanent_key($prod,$keyprodname,$key);
    # if $prod->{prod} is Enterprise and $prod->{prod} is not licensed
    # we need to ask user to input another permanent key
    if ($prod->{prod} =~ /Enterprise/mxi && $keyprodname !~ /Enterprise/mxi) {
        $msg=Msg::new("$keyprodname is licensed on all systems. Do you want to enter additional license?");
        while (1) {
            Msg::n();
            $ayn=$msg->aynn();
            if ($ayn eq 'Y') {
                ($keyprodname,$key)=$rel->ask_for_permanent_key($prod);
                next if (EDR::getmsgkey($keyprodname,'back'));
                $rel->register_permanent_key($prod,$keyprodname,$key);
            }
            # only ask for additional permanent key for one time
            last;
        }
    }
    return 1;
}

sub ask_for_permanent_key {
    my ($rel,$prod)=@_;
    my ($defaultopt,$help,$backopt);
    my ($msg,$key,$prodname);
    $backopt=1;
    Msg::n();
    # For Enterprise product, if the task is not upgrade, we allow user to input a non-Enterprise(smaller) key.
    if ($prod->{prod} =~ /Enterprise/mxi && !Cfg::opt(qw/upgrade patchupgrade hotfixupgrade/)) {
        $msg=Msg::new("Enter a valid license key:");
    } else {
        $msg=Msg::new("Enter a valid $prod->{abbr} license key:");
    }
    while (1) {
        $key=$msg->ask($defaultopt,$help,$backopt);
        return $key if (EDR::getmsgkey($key,'back'));
        $prodname=$rel->validate_permanent_key($key,$prod);
        last if ($prodname);
    }
    return ($prodname,$key);
}

sub register_permanent_key {
    my ($rel,$prod,$keyprodname,$key)=@_;
    my ($cpic,$cfg,$msg,$keylist);
    my ($exit,$result);
    $cfg=Obj::cfg();
    $cpic=Obj::cpic();
    $msg=Msg::new("Registering $keyprodname user-defined key $key on $prod->{name}");
    $msg->print();
    # Start registering $key on each system
    # For all the products, if the $key is already registered on the system, skip registering.
    # For enterprise, allow certain combinations, if not fit for the combinations, skip registering.
    for my $sys (@{$cpic->{systems}}) {
        $keylist=$rel->licensed_keys_sys($sys);
        if (EDRu::inarr($key,@{$keylist})) {
            $msg=Msg::new("$key is already registered on $sys->{sys}, skip registering this key on $sys->{sys}");
            $msg->print();
            next;
        }
        ($exit,$result)=$rel->register_key_sys($sys,$key);
        if ($exit) {
            $msg=Msg::new("Failed to register user-defined key $key on $sys->{sys}:\n$result");
        } else {
            $msg=Msg::new("Successfully registered user-defined key $key on $sys->{sys}.");
        }
        $msg->print;
    }
    Msg::n();
    $cfg->set_value("keys,license","push",$key) unless (EDRu::inarr($key,@{$cfg->{keys}{license}}));
    return 1;
}

sub register_key_sys {
    my ($rel,$sys,$key)=@_;
    my ($command,$result,$exit,$msg,@lines);
    # register $key
    $command = $rel->get_lic_binary_path("vxlicinstupgrade -k $key");
    $result=$sys->cmd("$command 2>/dev/null");
    # Remove the empty(space) lines
    $result=~s/^\s*\n//g;
    @lines=split(/\n/,$result);
    # remove the first 2 lines (licence version and license copyright)
    splice (@lines,0,2);
    $result=join('\n',@lines);
    # remove the empty lines in the array
    $exit=EDR::cmdexit();
    if ($exit) {
        return ($exit,$result);
    }
    # each time when a new key is registered, reset $sys->{donotquerykeys}
    undef ($sys->{donotquerykeys});
    return ($exit,$result);
}

# If any system has keyless key, return 1,
# If none of the systems have keyless key, return 0
sub keyless_key_exists {
    my ($rel)=@_;
    my ($cpic,$vxkeyless);
    $cpic=Obj::cpic();
    for my $sys (@{$cpic->{systems}}) {
        # for each system, check vxkeyless result
        $vxkeyless=$rel->get_vxkeyless_key_sys($sys);
        return 1 if ($vxkeyless);
    }
    return 0;
}

# If any system has pervious keyless key such as SFHASTD_60, return 1,
# If none of the systems have this kind of keyless key, return 0
sub previous_keyless_key_exists {
    my ($rel)=@_;
    my ($cpic,$vxkeyless);
    $cpic=Obj::cpic();
    for my $sys (@{$cpic->{systems}}) {
        for my $key (split(/,/,$sys->{vxkeyless})) {
            if ($key =~ /\d+/) {
                return 1;
            }
        }
    }
    return 0;
}

# If any system has permanent key with version <= $rel->{license_ver}, return 1
# If all systems' all keys have permanent key version = $rel->{license_ver}, reutrn 0
sub previous_permanent_key_exists {
    my ($rel)=@_;
    my ($cpic,$keys,$version);
    $cpic=Obj::cpic();
    for my $sys (@{$cpic->{systems}}) {
        for my $key (keys%{$sys->{keys}}) {
            $version=$sys->{keys}{$key}{Version};
            next unless ($version);
            # compare the version
            return 1 if (EDRu::compvers($rel->{license_ver},$version,2)==1);
        }
    }
    return 0;
}

# inquiry keyless keys
sub get_vxkeyless_key_sys {
    my ($rel,$sys)=@_;
    my ($command,$vxkeyless);

    $command = $rel->get_lic_binary_path("vxkeyless display");
    $vxkeyless=$sys->cmd("$command 2>/dev/null");
    return if ($vxkeyless =~ /No keys installed/mi);
    $sys->set_value('vxkeyless',$vxkeyless);
    return $vxkeyless;
}

# If any system has permanent key, return 1,
# If none of the systems have permanent key, return 0
# Do not treat single node CPS key as a normal permanent key
sub permanent_key_exists {
    my ($rel,$prod)=@_;
    my $cpic=Obj::cpic();
    for my $sys (@{$cpic->{systems}}) {
        if ($sys->{keys} && scalar(keys%{$sys->{keys}})>0) {
            return 0 if ($rel->is_single_node_cps_key_sys($sys));
            return 1;
        }
    }
    return 0;
}

# validate a provided $key on a provided $sys
# valid: return the parent product name of this key
# invalid: return 0;
# makeresponsefile option: return 1;
sub validate_permanent_key {
    my ($rel,$key,$prod) = @_;
    my ($msg,$edr,$cpic,$sys,$valid,$keyless,$prodname,$version,$vxlicrep,$command);
    my (@prodnames);
    return 1 if (Cfg::opt('makeresponsefile'));
    return 0 unless ($key);

    $edr=Obj::edr();
    $cpic=Obj::cpic();
    $sys=${$cpic->{systems}}[0];
    Msg::log("Validating $key");
    $command = $rel->get_lic_binary_path("vxlicrep -k $key");
    $vxlicrep=$sys->cmd("$command 2>/dev/null");
    return 0 unless ($vxlicrep);
    
    $valid=1;
    for my $line (split (/\n/, $vxlicrep)) {
        if ($line =~ /^\s*Product Name\s+=\s*(.+)\s*$/) {
            push(@prodnames,$1);
        } elsif (!$keyless && $line =~ /^\s*VXKEYLESS\s+=\s*(.+)\s*$/) {
            $keyless=$1;
        } elsif (!$version && $line =~ /^\s*Version\s+=\s*(.+)\s*$/) {
            $version=$1;
        }
    }
    # Retrieve the parent product name from @prodnames
    # the parent product name must include Enterprise/Storage/Availability/Foundation
    for my $name (@prodnames) {
        if ($name =~ /InfoScale Enterprise|InfoScale Storage|InfoScale Availability|InfoScale Foundation/mi) {
            $prodname=$name;
            $prodname=~s/\s*VERITAS\s*//gi;
            last;
        }
    }
    if (!@prodnames) {
        $msg=Msg::new("$key is an invalid key.");
        $valid=0;
    } elsif ($keyless) {
        $msg=Msg::new("$key is a keyless key. If you prefer keyless licensing, choose the keyless licensing option from the Licensing main menu.");
        $valid=0;
    } else {
        if ($version) {
            # Only compare 2 digits, in 7.0.1 release, Sol10x64 also compare 2 digits.
            if (EDRu::compvers($version,$rel->{license_ver})==2) {
                # The license key version is lower than $rel->{license_ver}
                $msg=Msg::new("Registering a license key with version lower than $rel->{license_ver} is not allowed.");
                $valid=0;
            } elsif (!$prodname) {
                # $prodname make sense only when $version >= $rel->{license_ver}
                $msg=Msg::new("Cannot detect the parent product from the license key: $key.");
                $valid=0;
            } else {
                # When the Product Name != Key
                # for Storage/Availability/Foundation, only the following scenarios are allowed:
                #            Product Name                                         Key
                #       $prod->{prod} = Storage/Availability/Foundation, $prodname = Enterprise
                #       $prod->{prod} = Foundation,                      $prodname = Storage
                #       $prod->{prod} = Enterprise,                      $prodname = Any
                # However, in upgrade case, only the following scenarios are allowed:
                #            Product Name                                         Key
                #       $prod->{prod} = Storage/Availability/Foundation, $prodname = Enterprise
                #       $prod->{prod} = Foundation,                      $prodname = Storage
                if ($prodname !~ /$prod->{prod}/mxi) {
                    Msg::log("The registered license is not the same with the product installed.");
                    if (Cfg::opt(qw/upgrade patchupgrade hotfixupgrade/)) {
                        if (($prod->{prod} =~ /Storage|Availability|Foundation/mxi && $prodname =~ /Enterprise/mix) || ($prod->{prod} =~ /Foundation/mix && $prodname =~ /Storage/mix)) {
                            Msg::log("Accepting the product name as $prod->{abbr} and the key to be registered is for $prodname");
                        } else {
                            $msg=Msg::new("$key is a license key for $prodname, it is not allowed to register $prodname license key on $prod->{name}.");
                            $valid=0;
                        }
                    } else {
                        if ($prod->{prod} =~ /Enterprise/mxi || ($prod->{prod} =~ /Storage|Availability|Foundation/mxi && $prodname =~ /Enterprise/mix) || ($prod->{prod} =~ /Foundation/mix && $prodname =~ /Storage/mix)) {
                            Msg::log("Accepting the product name as $prod->{abbr} and the key to be registered is for $prodname");
                        } else {
                            $msg=Msg::new("$key is a license key for $prodname, it is not allowed to register $prodname license key on $prod->{prod}.");
                            $valid=0;
                        }
                    }
                }
           }
        } else {
            $msg=Msg::new("Cannot detect version information from license key: $key.");
            $valid=0;
        }
    }
    if ($valid) {
        return $prodname;
    }
    $msg->bold();
    return 0;
}

# checking already installed keys on all the systems
# all the $sys->{keys} will be initiated after this function called
sub license_status {
    my ($rel) = @_;
    my ($edr,$cpic,$prod,$license_status);
    my (@vlic_installed,@vlic_uninstalled,@licensed,@unlicensed);
    
    $edr=Obj::edr();
    $cpic=Obj::cpic();
    $prod=$cpic->prod($cpic->{prod});
    
    for my $sys (@{$cpic->{systems}}) {
        # Need to force retrieve the license status on each system because 
        # the license detected with previous release VRTSvlic maybe incorrect.
        undef ($sys->{donotquerykeys}) if (Cfg::opt('upgrade'));
        # Check VRTSvlic status first.
        if ($rel->vlic_installed($sys)) {
            push(@vlic_installed, $sys->{sys});
            # Check if the target product is licensed or not
            if ($prod->licensed_sys($sys)) {
                push(@licensed, $sys->{sys});
            } else {
                push(@unlicensed, $sys->{sys});
            }
        } else {
            push(@vlic_uninstalled, $sys->{sys});
        }
    }

    $rel->{license_status} = {
        'vlic_installed'   => \@vlic_installed,
        'vlic_uninstalled' => \@vlic_uninstalled,
        'licensed'         => \@licensed,
        'unlicensed'       => \@unlicensed
    };

    return 1;
}

sub vlic_installed {
    my ($rel,$sys) =@_;
    my $pkg;
    $pkg=$sys->pkg('VRTSvlic32');
    return 1 if ($pkg->version_sys($sys));
    if (Obj::rel()->{simple}) {
        $pkg=$sys->pkg('VRTSinfr731');
        return 1 if ($pkg->version_sys($sys));
    }
    return 0;
}

sub prod_licensed_sys {
    my ($rel,$sys,$prodname,$skip_superset) = @_;
    my ($cpic,$msg,$pname,$lname,$ha_status,$vxkeyless,$vxk,$lk,$prod,$licensed,$vcs_lic,$sf_lic);
    return 1 if ($rel->is_single_node_cps_key());
    return 1 if (Cfg::opt(qw(makeresponsefile)) && !Cfg::opt('upgrade'));
    $cpic=Obj::cpic();
    $ha_status = 0;
    $skip_superset ||= 0;
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    if ($prodname) {
        # the following line is appha fix for 3541080
        # due to the inconsistency between $sys->{iprod} and $rel->{prods}
        # $sys->{iprod} and JSON files use 'appha'
        # $rel->{prods} uses APPLICATIONHA
        $prodname=~s/^APPHA/APPLICATIONHA/i;
        # The following lines are for sfcfs fix in 3564266
        # older releases may have sfcfs which is already gone in 6.2.
        # therefore need to take it as SFCFSHA for license version check
        # the same to SVS. Since SFCFS is considered a component exist in 7.0,
        # Only do the replace for SVS
        if ($prodname=~/^SVS\d+/i) {
            $prodname=~s/^SVS/SFCFSHA/i;
        }
        # The following lines are for sfsybce fix in 3865111
        # Version Check of SF SybaseCE will be set as product sfsybce
        if ($prodname=~/^SFSYBCE\d+/i) {
            $prodname=~s/^SFSYBCE/SFSYBASECE/i;
        }
        return 0 unless (EDRu::inarr($prodname,@{$rel->{prods}}));
    }
    $prod = ($prodname) ? $sys->prod($prodname) : $cpic->prod;
    $prodname||=$prod->{abbr};
    # read license into $sys->{$keys}
    $rel->read_licenses_sys($sys);
    
    if ((!$skip_superset) && $prod->{licsuperprods}) {
        for my $superprod (@{$prod->{licsuperprods}}) {
            $licensed=$rel->prod_licensed_sys($sys,$superprod);
            return $licensed if ($licensed);
        }
    }
    if ($sys->{vxkeyless}) {
        $vxkeyless=$rel->get_vxkeyless_prod_level($sys, $prod);
        $vxkeyless="SF" if ($prod->{abbr} eq 'SF');
        $lk='';
        # vxkeyless is the key that the $prod supposed to have
        # vxk is the key that's discovered in $sys
        for my $vxk (split /,/m,$sys->{vxkeyless}) {
            if ($vxkeyless eq $vxk) {
                $lk='vxkeyless';
                last;
            }
            if ($vxk =~ /^$vxkeyless(STD|STD_|ENT|ENT_|_|$)/mx) {
                $lk='vxkeyless';
                last;
            }
            # Under the following circumstances consider the product licensed.
            # Key on the System ($vxk)             Product Key Needed
            #   Enterprise                              Any  
            #   Storage                                 Foundation
            #   Any of the 4 key(X)                     Same key(X)
            if ($vxk =~ /^$vxkeyless/mx || $vxk=~/Enterprise/mxi || ($vxk=~/Storage/mxi && $vxkeyless=~/Foundation/mxi)) {
                $lk='vxkeyless';
                last;
            }
        }
        if ($lk) {
            #$vxk=$rel->get_vxkeyless_key_version($sys,$prod);
            #return "$vxk $lk";
            return $lk;
        }
    }
    for my $lk (sort { $a cmp $b } keys(%{$sys->{keys}})) {
        $pname=$sys->{keys}{$lk}{'Product Name'};
        $pname=~s/^VERITAS //mi;
        for my $lname (@{$prod->{lic_names}}) {
            $lname=~s/^VERITAS //mi;
            if ($pname=~/$lname/m) {
                if($prodname =~ /APPLICATIONHA/mi){
                    if($sys->{keys}{$lk}{ApplicationHA} =~ /Enabled/mi) {
                        return ($sys->{keys}{$lk}{Version}) ? $sys->{keys}{$lk}{Version} : 1;
                    }
                    next;
                }
                # for mutiple keys, return the highest version
                $licensed = $sys->{keys}{$lk}{Version} if (!$licensed || (EDRu::compvers($licensed,$sys->{keys}{$lk}{Version}) == 2));
                if ($prodname=~/SFHA/m) {
                    # the following line is really just for SFHA
                    $ha_status = $rel->feature_licensed_sys($sys, 'Mode#VERITAS Cluster Server', 'VCS');
                    $vcs_lic = $rel->prod_licensed_sys($sys,'VCS731');
                    $sf_lic = $rel->prod_licensed_sys($sys,'SF731',1);
                    $ha_status = $vcs_lic if (EDRu::compvers($ha_status,$vcs_lic) == 2);
                    $ha_status = $sf_lic  if ($ha_status && EDRu::compvers($ha_status,$sf_lic)  == 2);
                    $licensed = 0 if ($lname !~ /$pname/m);
                    $licensed = $ha_status if ($ha_status && (EDRu::compvers($licensed,$ha_status) == 2));
                }
                if (!$licensed && ($pname=~/InfoScale/mi)) {
                    # For 7.0 and later
                    $licensed=$sys->{keys}{$lk}{Version}||1;
                }
            }
        }
    }

    if ($prod->{abbr} eq 'SF') {
        # need to check whether the SFBasic license is the only key exists on the system for ETrack 3131566
        for my $lk (sort { $a cmp $b } keys(%{$sys->{keys}})) {
            $pname=$sys->{keys}{$lk}{'Product Name'};
            $pname=~s/^VERITAS //mi;
            next unless ($pname =~ /^Storage Foundation/);
            if ($pname eq "Storage Foundation Basic") {
                $sf_lic=1;
            } else {
                $sf_lic=0;
                last;
            }
        }
        if ($sf_lic) {
            $licensed.=" SFBasic";
            $prod->set_value('mode', "SF Basic");
            $prod->set_pkgs_sfbasic();
        }
    }

    return 1 if ($licensed);
    return 0;
}

# store keys found on a given $sys into $sys->{keys}
# after reading all the registered ones on $sys via vxlicrep_sys call
sub read_licenses_sys {
    my ($rel,$sys) = @_;
    my ($lictype,$feat,@f,$line,$vxlicrep,$val,$lk);
    return if ($rel->{nolic});
    # the function is used in many places, need to ensure vlic is installed
    return unless ($rel->vlic_installed($sys));
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    
    # if $sys->{donotquerykeys} set, skip check again
    return 0 if $sys->{donotquerykeys};
    delete $sys->{keys};
    
    $vxlicrep=$rel->vxlicrep_sys($sys);
    $rel->get_vxkeyless_key_sys($sys);
    @f=split(/\n+/,$vxlicrep);
    for my $line (@f) {
        next unless ($line=~/=/m);
        next if ($line=~/Features :=/m);
        $line=~s/^\s+//m;
        if ($line=~/License Key/m) {
           $lk++;
           # store license type in case of editions key
           # will reset if not
           $sys->set_value("keys,$lk,License Type", $lictype);
        }
        next if (!$lk);
        ($feat,$val)=split(/\s+=\s+/m,$line,2);
        # Trim the spaces
        $feat =~ s/^\s+|\s+$//g;
        $val  =~ s/^\s+|\s+$//g;
        if ($feat eq 'License Type') {
           $sys->set_value("keys,$lk,$feat", $val);
           $lictype = $val;
        } else {
           $sys->set_value("keys,$lk,$feat", $val) if (!($sys->{keys}{$lk}{$feat}));
        }
    }
    $sys->set_value('donotquerykeys', 1);
    return 1;
}

sub vxlicrep_sys {
    my ($rel,$sys) = @_;
    my ($cpic,$msg,$pdfr,$vxlicrep,$pkg,$result);
    my ($rootpath,$vxlic_path);
    return '' if (Cfg::opt('makeresponsefile') && Cfg::opt(qw(configure uninstall)));

    $cpic=Obj::cpic();
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return '' if ($sys->{stop_checks});
    return '' unless ($rel->vlic_installed($sys));
    
    # Check licenses on rootpath directory if rootpath is set.
    $rootpath = Cfg::opt('rootpath');
    $vxlic_path = "$rootpath/etc/vx/licenses/lic/";
    # rootpath not set, directly return the result of $vxlicrep
    unless ($rootpath && $rootpath ne '/') {
        $vxlicrep = $sys->cmd("_cmd_vxlicrep 2>/dev/null");
        return $vxlicrep;
    }
    # Check if license pkg is installed on PBE as we have to run vxlicrep cmd on PBE.
    # there will be duplicated result because each of the key used vxlicrep -k <key> and one of the key involves other keys.
    if ($sys->is_dir($vxlic_path)) {
        my $lics = $sys->cmd("cd $vxlic_path; _cmd_ls *.vxlic | _cmd_cut -d '.' -f 1 2>/dev/null");
        for my $lic (split (/\n/, $lics)) {
            $result=$sys->cmd("chroot $rootpath _cmd_vxlicrep -k $lic 2>/dev/null");
            # Due to incident 3598716
            # On Solaris 11 GA upgrade Solaris 11 Ux BE, the above command will fail
            # We assume Solaris 11 GA 99% has product installed, hence use local vxlicrep to try again
            if (EDR::cmdexit()) {
                $result=$sys->cmd("_cmd_vxlicrep -k $lic 2>/dev/null");
            }
            $vxlicrep.=$result;
        }
    }
    return $vxlicrep;
}

# Given a product, return the vxkeyless prod level registered on the $sys
# e.g. Give SF, it will return SF_VFR (when VFR is registered)
sub get_vxkeyless_prod_level {
    my ($rel,$sys,$prod,$prod_only) = @_;
    my ($new_prod,$uname);
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    return '' if ($sys->{stop_checks});
    $prod=$sys->prod($prod) unless (ref($prod));

    if ($prod->{prod} eq 'SF') {
        $new_prod=(Cfg::opt('prodmode') eq 'SF Basic') ? 'SFBASIC' :
                  (Cfg::opt('prodmode') eq 'SF Standard') ? 'SFSTD' :
                  (Cfg::opt('prodmode') eq 'SF Enterprise') ? 'SFENT' : 'SFSTD';
    } elsif ($prod->{prod} eq 'SFHA') {
        $new_prod=(Cfg::opt('prodmode') eq 'SF Standard HA') ? 'SFHASTD' :
                  (Cfg::opt('prodmode') eq 'SF Enterprise HA') ? 'SFHAENT' : 'SFHA';
    } elsif ($prod->{prod} =~ /^SF(CFS|CFSHA|CFSRAC|RAC)$/mx) {
        $new_prod=$prod->{prod} . 'ENT';
    } elsif($prod->{prod} =~ /^(FS|VM)$/m) {
        $new_prod='SFSTD';
    } elsif($prod->{prod} =~ /^ApplicationHA$/m) {
        $new_prod='APPLICATIONHA';
    } elsif($prod->{prod} =~ /^Foundation$/im) {
        $new_prod='FOUNDATION';
    } elsif($prod->{prod} =~ /^Storage$/im) {
        $new_prod='STORAGE';
    } elsif($prod->{prod} =~ /^Availability$/im) {
        $new_prod='AVAILABILITY';
    } elsif($prod->{prod} =~ /^Enterprise$/im) {
        $new_prod='ENTERPRISE';
    } else {
        $new_prod=$prod->{prod};
    }
    
    # Starting from 7.0 we don't have suffix for each product
    return $new_prod if ($new_prod=~/^(FOUNDATION|STORAGE|AVAILABILITY|ENTERPRISE)/mx);
    
    # If !$prod_only, need to add suffix like _VR, _GCO... if it applies
    if (!$prod_only) {
        # _VR & _VFR don't apply to VCS
        # _VFR doesn't apply to non-Linux
        if ($new_prod !~ /^VCS/) {
            $new_prod.='_VR' if (Cfg::opt('vr') || Cfg::opt('vvr'));
            # _VR includes _VFR, and they don't show up together
            $new_prod.='_VFR' if (Cfg::opt('vfr') && (!Cfg::opt('vr')) && ($new_prod !~ /^SVS/) && (Padv::plat($sys->{padv}) eq "Linux"));
        }
        $new_prod.='_GCO' if (Cfg::opt('gco') && ($new_prod =~ /^(VCS|SFHAENT|SFCFSHAENT|SFRACENT|SVS|SFSYBASECE)/mx));
        $new_prod='SFBASIC' if ($new_prod =~ /^SFBASIC/i);
        $new_prod.='_EVAL' if (Cfg::opt('eval'));
        # DMP has no '_<blah>' at all. Etrack 3265969
        $new_prod='DMP' if ($new_prod =~ /^DMP/);
    }

    return $new_prod;
}

# return 1 if a feature is ENABLED, Yes, or matches passed $val
sub feature_licensed_sys {
    my ($rel,$sys,$fm,$val) = @_;
    my ($feat,$fval,$lk,$version,$ver);
    return 1 if ($rel->is_single_node_cps_key());
    $rel->read_licenses_sys($sys);
    $version = 0;
    for my $lk (sort { $a cmp $b } keys(%{$sys->{keys}})) {
        for my $feat (sort { $a cmp $b } keys(%{$sys->{keys}{$lk}})) {
            $ver = 0;
            $fval = $sys->{keys}{$lk}{$feat};
            next unless (($feat =~ /^$fm$/i) || (($fm eq "VVR") && ($feat =~ /^VVR\#\w*/)));
            if (($fval eq 'Enabled') || ($fval eq 'YES')) {
                $ver = ($sys->{keys}{$lk}{Version} !~ /^\d+/) ? 1 : $sys->{keys}{$lk}{Version};
            }
            $version = $ver if ($ver && ($version < $ver));
            $ver = $sys->{keys}{$lk}{Version} if (($val) && ($fval eq $val));
            $version = $ver if ($ver && ($version < $ver));
        }
    }
    return $version;
}

sub vr_licensed_sys {
    my ($rel,$sys) = @_;
    $rel->read_licenses_sys($sys);
    if($rel->feature_licensed_sys($sys,'VVR')) {
        Cfg::set_opt('vr');
        return 1;
    }
    return '';
}

sub vfr_licensed_sys {
    my ($rel,$sys) = @_;
    $rel->read_licenses_sys($sys);
    # VFR can only be considered when VVR/VR is not available
    return '' if ($rel->vr_licensed_sys($sys));
    if($rel->feature_licensed_sys($sys,'VFR')) {
        Cfg::set_opt('vfr');
        return 1;
    }
    return '';
}

sub gco_licensed_sys {
    my ($rel,$sys) = @_;
    $rel->read_licenses_sys($sys);
    if($rel->feature_licensed_sys($sys,'Global Cluster Option')) {
        Cfg::set_opt('gco');
        return 1;
    }
    return '';
}



# The following functions from previous version are currently not been used.
# Since they're functional related, Keeping them for now.
sub create_vxlic_license_key_dir {
    my ($rel,$sys) = @_;
    my ($key_dir);

    $key_dir=$rel->get_lic_data_file_path("/etc/vx/licenses/lic");
    unless ($sys->is_dir($key_dir)) {
        if ($sys->mkdir($key_dir)) {
            $sys->cmd("_cmd_chmod 755 $key_dir");
            $sys->cmd("_cmd_chown root:root $key_dir");
        }
    }
    return '';
}

# return ref of array of registered keys on a given $sys
sub licensed_keys_sys {
    my ($rel,$sys) = @_;
    my (@prod_keys,$lname,$lk);

    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $rel->read_licenses_sys($sys);

    for my $lk (keys %{$sys->{keys}}) {
        push(@prod_keys,$sys->{keys}{$lk}{'License Key'});
    }
    return \@prod_keys;
}

# For SFBasic only
# return the content of .key file if found in the mediapath
sub read_sfbasic_key {
    my ($rel) = @_;
    my ($edr,$key);

    $edr=Obj::edr();
    $key=EDR::cmd_local("_cmd_find $edr->{mediapath} -type f -name .key 2>/dev/null");
    $key=~s/^\s+//mg;
    $key=(split /\n/, $key)[0];
    if ($key) {
        $key=EDRu::readfile($key);
        chomp $key;
    }
    return $key;
}

# return the prod vxkeyless level version
# the version should be $rel->{keyless_license_ver} if the key doesn't have _\d+ suffix
# otherwise the version is \d+
sub get_vxkeyless_key_version {
    my ($rel,$sys,$prod) = @_;
    my ($version,$key_stem);

    $key_stem=$prod->{prod};
    $key_stem="SF" if ($key_stem =~ /^(FS|VM)$/m) ;

    for my $key (split /,/, $sys->{vxkeyless}) {
        # the keyless key must be able to enable the product($key_steam)
        next unless ($rel->vxkeyless_keys_inclusion($key,$key_stem));
        $version=$rel->{keyless_license_ver};
        last unless ($key =~ /_\d+/);
        # assuming all the old release keys are like VCS_5.1SP1, VCS_6.2
        $version=(split /_/, $key)[-1];
    }

    $version||=$rel->{keyless_license_ver};
    return $version;
}

# determine whether $key1 enables $key2
sub vxkeyless_keys_inclusion {
    my ($rel,$key1,$key2)=@_;

    if ($key2 eq 'DMP') {
        # All SF-related keys enable DMP
        return 1 if ($key1 =~ /^(SF|DMP|SVS)/);
    } elsif ($key2 eq 'VCS') {
        return 1 if ($key1 =~ /^(SFHA|VCS|SFCFS|SFSYB|SFRAC|SVS)/);
    } elsif ($key2 eq 'SF') {
        return 1 if ($key1 =~ /^(SF|SVS)/);
    } elsif ($key2 eq 'SFHA') {
        return 1 if ($key1 =~ /^(SFHA|SFCFS|SFSYB|SFRAC|SVS)/);
    } elsif ($key2 eq 'SFCFS') {
        return 1 if ($key1 =~ /^(SFCFS|SFSYB|SFRAC|SVS)/);
    } elsif ($key2 eq 'SFCFSHA') {
        return 1 if ($key1 =~ /^(SFCFSHA|SFSYB|SFRAC|SVS)/);
    } elsif ($key2 eq 'SVS') {
        # SVS is merged into SFCFSHA in 6.1
        return 1 if ($key1 =~ /^(SFCFSHA|SVS)/);
    } elsif ($key2 =~ /^SFSYB/) {
        return 1 if ($key1 =~ /^SFSYB/);
    } elsif ($key2 eq 'SFRAC') {
        return 1 if ($key1 =~ /^SFRAC/);
    } elsif ($key2 eq 'APPLICATIONHA') {
        # Right now 6.1 VCS keyless key doesn't enable ApplicationHA
        # return 1 if ($key1 =~ /^(SFHA|VCS|SFCFS|SFSYB|SFRAC|SVS)/);
        return 1 if ($key1 =~ /^APPLICATIONHA/);
    }
    
    # Starting from 7.0 there can only be one keyless key (no overlapping)
    if ($key1 eq $key2) {
        return 1;
    }
    return 0;
}

# return the key type on a given set of @$sitelic
# or @$lictype
sub keytype {
    my ($rel,$lictype,$sitelic) = @_;
    my ($cpic,$hi,@kt,$lt,$type,$prod);
    $cpic=Obj::cpic();
    $prod=$cpic->prod;
    for my $lt (@$sitelic) {
        return Msg::get('site') if ($lt=~/YES/m);
    }
    @kt=qw(UNKNOWN DEMO EVAL PERMANENT SITE);
    $hi=0;
    for my $lt (@$lictype) {
        $type = ($lt=~/SITE/m) ? 4 : ($lt=~/PERMANENT/m) ? 3 :
                ($lt=~/EVAL/m) ? 2 : ($lt=~/DEMO/m) ? 1 : 0;
        $hi=$type if ($type>$hi);
    }
    return ($kt[$hi]=~/^PABBR$/m) ? $prod->{prod} : Msg::get("$kt[$hi]");
}

# return a list of feature values for a specific product
# returns a list because multiple keys may provide multiple values
sub prod_feature_values_sys {
    my ($rel,$sys,$prodname,$feat) = @_;
    my ($cpic,$pname,$fv,$lname,$rv,$lk,$prod);
    $cpic=Obj::cpic();
    $prod = ($prodname) ? $sys->prod($prodname) : $cpic->prod;
    $rv=[];
    for my $lk (sort { $a cmp $b } keys(%{$sys->{keys}})) {
        $pname=$sys->{keys}{$lk}{'Product Name'};
        $pname=~s/^VERITAS //mi;
        for my $lname (@{$prod->{lic_names}}) {
            $lname=~s/^VERITAS //mi;
            if ($pname=~/$lname/m) {
               $fv=$sys->{keys}{$lk}{$feat};
               push(@$rv,$fv) if ($fv);
            }
        }
    }
    return $rv;
}

# return a list of feature values
# returns a list because multiple keys may provide multiple values
sub feature_values_sys {
    my ($rel,$sys,$fm) = @_;
    my ($feat,$fval,$rv,$val,$lk);
    $rv=[];
    $rel->read_licenses_sys($sys);
    for my $lk (sort { $a cmp $b } keys(%{$sys->{keys}})) {
        for my $feat (keys(%{$sys->{keys}{$lk}})) {
            $fval=$sys->{keys}{$lk}{$feat};
            push(@$rv,$fval) if ($feat eq $fm);
        }
    }
    return $rv;
}

# separated from license_sys, validating a provided key on the provided $sys
sub license_special_key {
    my ($rel,$sys) = @_;
    my ($cpic,$msg,$sitelic,$cfg,$lictype,$keytype,$prodname,$licupi,$prod);

    return '' if (!$sys);
    $cpic=Obj::cpic();
    $prod=$cpic->prod;
    $prodname=$prod->{abbr};

    if (!$sys->{special_lic}) {
        # display key info
        $licupi=$prod->{licupi};
        $cfg=Obj::cfg();
        $licupi||=$cfg->{upi};
        $lictype=prod_feature_values_sys($cpic,$sys,$licupi,'License Type');
        $sitelic=prod_feature_values_sys($cpic,$sys,$licupi,'Site License');
        $keytype=keytype($cpic,$lictype,$sitelic);
        return '' if (!$keytype);
        $msg=Msg::new("$keytype $prodname license registered on $sys->{sys}");
        $msg->print;
    }
    return;
}

# get all VRTSvlid defined vxkeyless keys on a provided $sys
sub get_all_valid_vxkeyless_keys {
    my ($rel,$sys) = @_;
    my ($keys,@lines,@all_keys,$command);

    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');

    if ($rel->vlic_installed($sys)) {
        $command = $rel->get_lic_binary_path("vxkeyless displayall");
        $keys=$sys->cmd("$command 2>/dev/null");
        @lines=split(/\n/, $keys);
        # remove the title line
        shift(@lines);
        for my $key (@lines) {
            if ($key =~ /^(\S+)\s*/m) {
                push (@all_keys, $1);
            }
        }
    }
    return \@all_keys;
}

# get the highest vxkeyless key already defined on the systems.
sub get_installed_prod_vxkeyless_highest_key {
    my ($rel)= @_;
    my($cpic,$prod,$cluster_nodes,$one_key_on_all_sys,%prod_keys,
       $key_stem,$key_highest,$key,$sys,$vr_option,$vfr_option,$gco_option,
       $eval_option,$std_or_ent,$nokeys);

    $cpic=Obj::cpic();
    $prod=$cpic->prod;

    # check the vxkeyless keys on all the systems.
    # if none of them have the same key for the prod, then
    # find out the highest product key and set it onto all systems
    $key_stem=$rel->get_vxkeyless_prod_level(${$cpic->{systems}}[0],$prod,1);
    $one_key_on_all_sys = 0;
    $vr_option = $vfr_option = $gco_option = $eval_option = $std_or_ent = 0;
    $std_or_ent = ($key_stem=~/STD$/m) ? 'STD' :
                  ($key_stem=~/ENT$/m) ? 'ENT' : 0;
    $key_stem=~s/(STD|ENT)$//mg;

    # set %prod_keys to all the $prod related keys,
    # including <PROD>{,HA}{,STD,ENT}(_VR|_VFR)(_GCO)(_EVAL)
    for my $sys (@{$cpic->{systems}}) {
        for my $key (sort (split /,/m,$sys->{vxkeyless})) {
            next if ($key !~ /^$key_stem/m);
            $prod_keys{$key}++;
        }
    }

    # assuming $cpic->{systems} has no duplicated sys
    # otherwise we're in trouble here
    # what we do below is to set the highest key to
    # the longest key that's included on all systems
    # of course, the _EVAL has the least privilege, if _VR|_VFR|_GCO exists
    $cluster_nodes=$#{$cpic->{systems}}+1;
    $key_highest = '';
    $nokeys = 1;
    for my $key (sort keys %prod_keys) {
        $nokeys = 0;
        # found a key exists on all systems
        if ($prod_keys{$key} == $cluster_nodes) {
            $one_key_on_all_sys = 1;
            # set the highest key to $key if
            #     no highest key set yet or
            #     highest key set already and the current new key doesn't have _EVAL suffix  or
            #     highest key has no ENT suffix and the current ney key has STD set
            #     any other possiblity here?
            $key_highest = $key if ((!$key_highest)   ||
                                    ($key !~ /_EVAL$/m) || ($key !~ /_EVAL_\d+/m) ||
                                    (($key_highest !~ /ENT/m) && ($key =~ /STD/m)));
            next;
        }
        # the following are for $key isn't found on all sys
        if ($key =~ /ENT/m) {
            $std_or_ent = 'ENT';
        } elsif ($key =~ /STD/m) {
            $std_or_ent = 'STD' if ($std_or_ent !~ /ENT/m);;
        } else {
            $std_or_ent = '' if ($std_or_ent !~ /(STD|ENT)/m);;
        }
        if ($key =~ /_VR|_VVR/m) {
            $vr_option = 1;
            $eval_option = 0;
        }
        if ($key =~ /_VFR/m) {
            $vfr_option = 1;
            $eval_option = 0;
        }
        if ($key =~ /_GCO/m) {
            $gco_option = 1;
            $eval_option = 0;
        }
        $eval_option = 1 if (($key =~ /_EVAL/m) && ! ($vr_option || $vfr_option || $gco_option));
    }

    if ($nokeys) {
        $key_highest=$rel->get_vxkeyless_prod_level(${$cpic->{systems}}[0],$prod,0) if (Cfg::opt(qw(license)));
    } elsif (!$one_key_on_all_sys) {
        # if no common prod key found on all sys
        # then we conbine the _VR|_VFR|_GCO found on all sys to form the highest key
        # _EVAL won't be taken if one of VR|GCO exists
        $key_highest = $key_stem;
        $key_highest .= "$std_or_ent" if ($std_or_ent);
        $key_highest .= '_VR' if ($vr_option || Cfg::opt('vr') || Cfg::opt('vvr'));
        $key_highest .= '_VFR' if ($vfr_option || Cfg::opt('vfr'));
        $key_highest .= '_GCO' if ($gco_option || Cfg::opt('gco'));
        $key_highest .= '_EVAL' if ($eval_option || Cfg::opt('eval'));
        $key_highest=$rel->get_vxkeyless_prod_level(${$cpic->{systems}}[0],$prod,0) if (Cfg::opt(qw(license)));
    }

    return $key_highest;
}

# check if the product has a permanent license (not keyless)
sub prod_permanent_licensed_sys {
    my ($rel,$sys,$prodi) = @_;
    my (@keyless_keys,@product_keys,$cpic,$exploded_keys,$lic,$lk,$lname,$pname,$prod,$temp_name,$command);
    return 1 if (Cfg::opt(qw(makeresponsefile)) && !Cfg::opt('upgrade'));
    $cpic=Obj::cpic();
    $sys=Obj::sys($sys) unless (ref($sys) eq 'Sys');
    $prod = ($prodi) ? $sys->prod($prodi) : $cpic->prod;
    $prodi ||= $prod->{abbr};
    $rel->read_licenses_sys($sys);
    for my $lk (sort { $a cmp $b } keys(%{$sys->{keys}})) {
        # skip APPHA key or CPS key for vcs product
        next if (($prodi =~ /VCS/) && (($sys->{keys}{$lk}{'ApplicationHA'} eq 'Enabled') ||
            ($sys->{keys}{$lk}{'Coordination Point Server'} eq 'Enabled')));
        # record all the keyless keys.
        $lic = $sys->{keys}{$lk}{'License Key'};
        if ($sys->{keys}{$lk}{'VXKEYLESS'} eq 'Enabled') {
            # get the list of exploded key of the keyless key
            $command = $rel->get_lic_binary_path("vxlicrep -k $lic");
            $exploded_keys = $sys->cmd("$command 2> /dev/null | _cmd_grep 'License Key'");
            for my $kk (split(/\n/, $exploded_keys)) {
                $kk =~ s/License Key\s+=//m;
                $kk = EDRu::despace($kk);
                push (@keyless_keys,$kk);
            }
        }
        $pname=$sys->{keys}{$lk}{'Product Name'};
        $pname=~s/^VERITAS //mi;
        for my $lname (@{$prod->{lic_names}}) {
            $temp_name = $lname;
            $temp_name=~s/^VERITAS //mi;
            if ($pname=~/$temp_name/m) {
                push (@product_keys,$lic);
            }
        }
    }
    @product_keys = @{EDRu::arrdel(\@product_keys, @keyless_keys)};
    return 1 if (@product_keys > 0);

    return 0;
}

sub get_lic_data_file_path {
    my ($rel,$datafile)=@_;
    my $rootpath;

    $datafile||="/etc/vx/licenses/dat/licenses.dat";

    $rootpath = Cfg::opt('rootpath');
    if ($rootpath && $rootpath ne '/') {
        $rootpath =~ s/\/*$//g;
        $datafile = $rootpath.$datafile;
    }
    return $datafile;
}

sub get_lic_binary_path {
    my ($rel,$binary)=@_;
    my $rootpath;

    if ($binary!~/^\//) {
        $binary="/opt/VRTSvlic/bin/$binary";
    }

    $rootpath = Cfg::opt('rootpath');
    if ($rootpath && $rootpath ne '/') {
        $binary="chroot $rootpath $binary";
    }
    return $binary;
}

package Rel::UXRT731::Sol11x64;
@Rel::UXRT731::Sol11x64::ISA = qw(Rel::UXRT731::Sol11sparc);

# on Solaris 11 x64 platform, partial upgrade is supported
# We need to support the co-existance of 7.0 && previous key
# Since only Availability supports Solaris x64 platform
# We check if the latest Availability permanent key exists
# If so, we skip reporting previous license key exists
# If not, we use the general logic
sub previous_permanent_key_exists {
    my ($rel)=@_;
    my ($cpic,$keys,$version);
    $cpic=Obj::cpic();
    for my $sys (@{$cpic->{systems}}) {
        for my $key (keys%{$sys->{keys}}) {
            if (($sys->{keys}{$key}{'Product Name'} eq $cpic->prod->{name})
            && EDRu::compvers($rel->{license_ver},$sys->{keys}{$key}{Version},2) == 0) {
                # The product is Availability and the product major version is the same with key version
                return 0;
            }
        }
    }
    return $rel->SUPER::previous_permanent_key_exists();
}

package Rel::UXRT731::Sol12x64;
@Rel::UXRT731::Sol12x64::ISA = qw(Rel::UXRT731::Sol11x64);

1;

