#
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC. All rights reserved.
# 
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF VERITAS
# TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE
# PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
# 
# The Licensed Software  and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software  -  Restricted Rights"
# and DFARS 227.7202,  Rights  in  "Commercial Computer Software or  Commercial
# Computer Software Documentation" as applicable, and any successor regulations,
# whether  delivered  by  Veritas  as  on premises or hosted services.  Any use,
# modification, reproduction release, performance, display or disclosure of the
# Licensed Software and Documentation by the U.S. Government shall be solely in
# accordance with the terms of this Agreement. $
#
use Macros;
use Module::facter::environment;
use Module::facter::hardware;
use Module::facter::network;
use Module::facter::operatingsystem;
use Module::facter::software;
use Module::facter::veritas;

use Module::system::authorized_key;
use Module::system::copy;
use Module::system::cron;
use Module::system::file;
use Module::system::get_url;
use Module::system::group;
use Module::system::lineinfile;
use Module::system::mount;
use Module::system::service;
use Module::system::shell;
use Module::system::template;
use Module::system::user;
use Module::system::yum;

use Module::veritas::uuid;
use Module::veritas::llt;
use Module::veritas::gab;
use Module::veritas::vxvm_diskgroup;
use Module::veritas::vxvm_volume;
use Module::veritas::vcs_config;
use Module::veritas::sfcfsha_config;
use Module::veritas::process;
use Module::veritas::vcs_servicegroup;
use Module::veritas::gms;
use Module::veritas::amf;
use Module::veritas::had;
use Module::veritas::odm;
use Module::veritas::vxfen;

package Module;
@Module::ISA = qw(Obj);

use strict;
use JSON;
use File::Basename;

sub init {
    my ($self,%opts) = @_;

    # Define directories for additional modules.
    $self->{module_dir} = [];

    # Define default modules
    $self->{modules} = [ qw(
        facter::environment
        facter::hardware
        facter::network
        facter::operatingsystem
        facter::software
        facter::veritas
        system::authorized_key
        system::copy
        system::cron
        system::file
        system::get_url
        system::group
        system::lineinfile
        system::mount
        system::service
        system::shell
        system::template
        system::user
        system::yum
        veritas::uuid
        veritas::llt
        veritas::gab
        veritas::vxvm_diskgroup
        veritas::vxvm_volume
        veritas::vcs_config
		veritas::sfcfsha_config
		veritas::process
        veritas::vcs_servicegroup
        veritas::gms
        veritas::amf
        veritas::had
        veritas::odm
        veritas::vxfen
    ) ];

    return $self;
}

# Public subroutines

sub is_supported_module {
    my ($self,$module)=@_;

    for my $item (@{$self->{modules}}) {
        return $item if ($item=~/^$module$/i);
    }
    return;
}

sub append_module_dir {
    my ($self,$dirs) = @_;
    my ($files,$module);

    for my $dir (split(/[,;]/, $dirs)) {
        next if (! -d $dir);

        push @{$self->{module_dir}}, $dir;

        $files=`cd $dir; find *.pm -type f 2>/dev/null`;
        for my $file (split(/\n+/, $files)) {
            eval { require $file; };
            if ($@) {
                Msg::log("Failed to require the module file '$file'");
                next;
            }

            $module=$file;
            $module=~s/\.pm$//;
            $module=~s/\//\::/;
            $module="$module";
            push @{$self->{modules}}, $module;
        }
    }

    return 1;
}

sub get_module_full_name {
    my ($self,$module) = @_;

    if ($module !~ /\::/) {
        for my $fullname (@{$self->{modules}}) {
            if ($fullname=~/\::${module}$/) {
                return $fullname;
            }
        }
    }
    return $module;
}

sub get_common_scripts {
    my ($self,$sys) = @_;
    my $localsys = Obj::localsys();
	my $common_scripts="";
	my $source_dir=dirname($INC{"Module.pm"});
    my $output=$localsys->cmd("_cmd_find $source_dir| _cmd_grep pm | _cmd_grep -v Module");
	my @files=split("\n",$output);
	for my $file(@files){
		my $package_scripts=EDRu::readfile($file);
            chomp $package_scripts;
            $package_scripts=~s/^1;\s*$//m;
            if ($package_scripts) {
                $common_scripts.=$package_scripts . "\n";
			}
	}
    return $common_scripts;
}

sub get_module_scripts {
    my ($self,$modules) = @_;
    my ($all_scripts,$module_scripts,$modulefile,$modulepath);

    $modules ||= $self->{modules};

    $all_scripts='';

    for my $module (@{$modules}) {

        $modulefile="Module/$module.pm";
        $modulefile=~s/\::/\//g;

        $modulepath=$INC{$modulefile};
        if (!defined $modulepath) {
            # the module was not statically loaded
            ($modulepath) = grep { -f $_ }
                                map { "$_/$modulefile" }
                                    @{ $self->{module_dir} }, @INC;
        }

        if (!defined $modulepath) {
            $modulefile=~s/Module\///;
            # the module was already statically loaded
            $modulepath=$INC{$modulefile};
        }

        if (!defined $modulepath) {
            # the module was not statically loaded
            ($modulepath) = grep { -f $_ }
                                map { "$_/$modulefile" }
                                    @{ $self->{module_dir} }, @INC;
        }
        if (!defined $modulepath) {
            Msg::log("Couldn't find module file for module '$module'");
        } else {
            $module_scripts=EDRu::readfile($modulepath);
            chomp $module_scripts;
            $module_scripts=~s/^1;\s*$//m;
            if ($module_scripts) {
                $all_scripts.=$module_scripts . "\n";
            }
        }
    }

    return $all_scripts;
}

sub get_main_scripts_sys {
    my ($self,$sys)=@_;

    my $main_scripts = << '_EOF_';
package main;

use strict;

my ($sys,$plat,$padv,$modules,$arg,$all_result,$module_result,$ref);

$plat='__PLATFORM__';
$padv='__PADV__';

$sys=Sys::__PADV__->new();
$sys->{plat}=$plat || 'UNKNOWN';
$sys->{padv}=$padv || 'UNKNOWN';

$modules=$ARGV[0] || '';
$arg=$ARGV[1] || '';

$all_result={};

for my $module (split(/[,;]/, $modules)) {

    $module_result='';
    if ($padv && "Module::${module}::${padv}"->can('execute_sys')) {
        $module_result=eval { "Module::${module}::${padv}"->execute_sys($sys,$arg) };
    } elsif ($plat && "Module::${module}::${plat}"->can('execute_sys')) {
        $module_result=eval { "Module::${module}::${plat}"->execute_sys($sys,$arg) };
    } elsif ("Module::${module}"->can('execute_sys')) {
        $module_result=eval { "Module::${module}"->execute_sys($sys,$arg) };
    } elsif ($padv && "${module}::${padv}"->can('execute_sys')) {
        $module_result=eval { "${module}::${padv}"->execute_sys($sys,$arg) };
    } elsif ($plat && "${module}::${plat}"->can('execute_sys')) {
        $module_result=eval { "${module}::${plat}"->execute_sys($sys,$arg) };
    } elsif ("${module}"->can('execute_sys')) {
        $module_result=eval { "${module}"->execute_sys($sys,$arg) };
    }

    next unless ($module_result);

    $ref=ref($module_result);
    if ($ref eq 'HASH') {
        for my $key (keys %{$module_result}) {
            $all_result->{$key}=$module_result->{$key};
        }
    } else {
        # ARRAY or SCALAR
        $all_result->{$module}=$module_result;
    }
    $sys->summary($module, $all_result);
}

if (keys %{$all_result}) {
    eval "use Data::Dumper";
    if ($@) {
        print "\$results=" . EDRu::hash2string($all_result) . ";";
    } else {
        $Data::Dumper::Sortkeys=1;
        $Data::Dumper::Indent=1;
        print Data::Dumper->Dump([$all_result], [qw(results)]);
    }
}

_EOF_

    $main_scripts=~s/__PLATFORM__/$sys->{plat}/g;
    $main_scripts=~s/__PADV__/$sys->{padv}/g;

    return $main_scripts;
}

sub get_all_scripts_sys {
    my ($self,$sys)=@_;
    my ($all_scripts,$perl_path);

    $perl_path=$sys->{perl_path};
    return if (!defined $perl_path);

    $all_scripts="#!$perl_path\n";
    $all_scripts.=$self->get_common_scripts($sys);
    $all_scripts.=$self->get_module_scripts();
    $all_scripts.=$self->get_main_scripts_sys($sys);

    return $all_scripts;
}

sub execute_module_locally_sys {
    my ($self,$sys,$module,$arg) = @_;
    my ($plat,$padv,$modulename);

    $plat=$sys->{plat};
    $padv=$sys->{padv};

    $modulename = "Module::$module::$padv";
    if ($padv && "${modulename}"->can('execute_sys')) {
        return "${modulename}"->execute_sys($sys,$arg);
    }
    $modulename = "Module::$module::$plat";
    if ($padv && "${modulename}"->can('execute_sys')) {
        return "${modulename}"->execute_sys($sys,$arg);
    }
    $modulename = "Module::$module";
    if ($padv && "${modulename}"->can('execute_sys')) {
        return "${modulename}"->execute_sys($sys,$arg);
    }

    $modulename = "$module::$padv";
    if ($padv && "${modulename}"->can('execute_sys')) {
        return "${modulename}"->execute_sys($sys);
    }
    $modulename = "$module::$plat";
    if ($padv && "${modulename}"->can('execute_sys')) {
        return "${modulename}"->execute_sys($sys);
    }
    $modulename = "$module";
    if ($padv && "${modulename}"->can('execute_sys')) {
        return "${modulename}"->execute_sys($sys);
    }
    return;
}

sub execute_modules_sys {
    my ($self,$sys,$modules,$arg) = @_;
    my (@valid_modules,$scripts,$scripts_file,$all_result,$module_result,$ref,$cpic);

    @valid_modules=();
	$cpic=Obj::cpic();

    if (!$modules || !@$modules) {
        push @valid_modules, @{$self->{modules}};
    } else {
        for my $module (@{$modules}) {
            $module=$self->get_module_full_name($module);
            if (EDRu::inarr($module, @{$self->{modules}}) &&
                !EDRu::inarr($module, @valid_modules)) {
                push @valid_modules, $module;
            } else {
                Msg::log("$module is not a supported module!");
            }
        }
    }

    return if (!@valid_modules);

    # Try to generate scripts for all modules
    if (!$sys->{modules_scripts_file}) {
        #$sys->{perl_path}='/usr/bin/perl';
        $scripts=$self->get_all_scripts_sys($sys);
        if ($scripts) {
            $scripts_file=EDR::get('tmpdir') . '/modules_scripts.pl';
            $sys->writefile($scripts, $scripts_file); 
            $sys->chmod($scripts_file, '0755');
            $sys->set_value('modules_scripts_file', $scripts_file);
        }
		if(Cfg::opt('f')){
			$sys->cmd("_cmd_cp $scripts_file /tmp/modules_scripts.pl");
			print ("the script file location is /tmp/modules_scripts.pl");
			return;
		}
    }

    $all_result={};

    if ($sys->{modules_scripts_file}) {
        my $modules1=join ',', @valid_modules;
        my $command="$sys->{modules_scripts_file} $modules1";
        $command.=" \'$arg\'" if ($arg);
        $module_result=$sys->cmd("$command");
		if(Cfg::opt('module') || Cfg::opt('facter')){
			$cpic->copy_module_log_files($sys);
		}
        #print "$module_result\n";
        my $results={};
        eval $module_result if ($module_result);
        $all_result= $results;
    } else {
        # If could not generate scripts, then perform modules discovery one by one
        for my $module (@valid_modules) {
            $module_result=$self->execute_module_locally_sys($sys,$module,$arg);
            next if (!defined $module_result);

            $ref=ref($module_result);
            if ($ref eq 'HASH') {
                for my $key (keys %{$module_result}) {
                    $all_result->{$key}=$module_result->{$key};
                }
            } else {
                # ARRAY or SCALAR
                $all_result->{$module}=$module_result;
            }
        }
    }

    return $all_result;
}

sub load_facts_sys {
    my ($self,$sys,$modules) = @_;
    my (@all_modules,@facter_modules,$results);

    if ($modules) {
       @facter_modules=split(/,/,$modules);
    } else {
       @all_modules=@{$self->{modules}};
       @facter_modules=grep { /^facter\::/ } @all_modules;
    }

    $results=$self->execute_modules_sys($sys, \@facter_modules);
    if (keys %{$results}) {
        $sys->set_hash_value('facts', $results);
    }
    return $sys->{facts};
}

sub dump_facts_sys {
    my ($self,$sys) = @_;

    my $facts=$sys->{facts};
    return 0 unless (defined $facts);

    print to_json($facts,{pretty=>1,canonical=>1});

    return 1;
}

1;
