#
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC. All rights reserved.
# 
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF VERITAS
# TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE
# PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
# 
# The Licensed Software  and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software  -  Restricted Rights"
# and DFARS 227.7202,  Rights  in  "Commercial Computer Software or  Commercial
# Computer Software Documentation" as applicable, and any successor regulations,
# whether  delivered  by  Veritas  as  on premises or hosted services.  Any use,
# modification, reproduction release, performance, display or disclosure of the
# Licensed Software and Documentation by the U.S. Government shall be solely in
# accordance with the terms of this Agreement. $
#

package Module::facter::network;

# Get the following facts for operating system
#    network_interfaces
#    network_default_ipv4_address
#    network_default_ipv4_interface
#    network_default_ipv4_gateway
#    network_default_ipv6_address
#    network_default_ipv6_interface
#    network_default_ipv6_gateway


package Module::facter::network::Linux;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts,$if_class_path,$ip_bin_path,$rc,$output,$content,@words,$word,$i,$len);
    my (@all_ipv4_addresses,@all_ipv6_addresses,$if_path,$file,$key,$type);

    $facts={};

    $ip_bin_path='_cmd_ip';
    $if_class_path='/sys/class/net';

    # get default ipv4 interface
    # Use the commands:
    #     ip -4 route get 8.8.8.8                     -> Google public DNS
    #     ip -6 route get 2404:6800:400a:800::1012    -> ipv6.google.com
    # to find out the default outgoing interface, address, and gateway
    ($rc,$output)=$sys->cmd("${ip_bin_path} -4 route get 8.8.8.8 2>/dev/null");
    if (!$rc && $output && $output!~/^unreachable/) {
        @words=split /\s+/, $output;
        $i=0;
        $len=scalar @words;
        $len--;
        while ($i<$len) {
             my $word=$words[$i];
             if ($word eq 'via') {
                 $i++;
                 $facts->{network_default_ipv4_gateway}=$words[$i];
             } elsif ($word eq 'dev') {
                 $i++;
                 $facts->{network_default_ipv4_interface}=$words[$i];
             } elsif ($word eq 'src') {
                 $i++;
                 $facts->{network_default_ipv4_address}=$words[$i];
             }
             $i++;
        }
    }

    # get default ipv6 interface
    ($rc,$output)=$sys->cmd("${ip_bin_path} -6 route get 2404:6800:400a:800::1012 2>/dev/null");
    if (!$rc && $output && $output!~/^unreachable/) {
        @words=split /\s+/, $output;
        $i=0;
        $len=scalar @words;
        $len--;
        while ($i<$len) {
             my $word=$words[$i];
             if ($word eq 'via') {
                 $i++;
                 $facts->{network_default_ipv6_gateway}=$words[$i];
             } elsif ($word eq 'dev') {
                 $i++;
                 $facts->{network_default_ipv6_interface}=$words[$i];
             } elsif ($word eq 'src') {
                 $i++;
                 $facts->{network_default_ipv6_address}=$words[$i];
             }
             $i++;
        }
    }

    # get all network interfaces
    ($rc,$output)=$sys->cmd("_cmd_ls ${if_class_path}/ 2>/dev/null");
    if (!$rc && $output) {
        for my $interface (split(/\s+/,$output)) {
            next if ($interface eq '.' || $interface eq '..');

            $if_path="${if_class_path}/$interface";

            # get all IPv4 and IPv6 address
            @all_ipv4_addresses=();
            @all_ipv6_addresses=();
            ($rc,$output)=$sys->cmd("${ip_bin_path} -o addr show dev $interface 2>/dev/null");
            if (!$rc && $output) {
                for my $line (split(/\n+/,$output)) {
                    if ($line=~/inet\s+(\S+)/m) {
                        push @all_ipv4_addresses, $1;
                    } elsif ($line=~/inet6\s+(\S+)/m) {
                        push @all_ipv6_addresses, $1;
                    }
                }
                $facts->{network_interfaces}{$interface}{ipv4_addresses}=[ @all_ipv4_addresses ] if (@all_ipv4_addresses);
                $facts->{network_interfaces}{$interface}{ipv6_addresses}=[ @all_ipv6_addresses ] if (@all_ipv6_addresses);
            }

            # get interface macaddress
            $file="${if_path}/address";
            $content=$sys->readfile($file);
            chomp $content;
            $facts->{network_interfaces}{$interface}{macaddress}=$content if ($content && $content ne '00:00:00:00:00:00');

            # get interface state
            $file="${if_path}/operstate";
            $content=$sys->readfile($file);
            chomp $content;
            $facts->{network_interfaces}{$interface}{active}='true' if ($content && $content eq 'up');

            # get interface mtu
            $file="${if_path}/mtu";
            $content=$sys->readfile($file);
            chomp $content;
            $facts->{network_interfaces}{$interface}{mtu}=$content if ($content);

            # get interface speed
            $file="${if_path}/speed";
            $content=$sys->readfile($file);
            chomp $content;
            $facts->{network_interfaces}{$interface}{speed}=$content if ($content);

            # get interface type
            $file="${if_path}/type";
            $content=$sys->readfile($file);
            chomp $content;
            $type='';
            if ($content eq '1') {
                $type='ether';
            } elsif ($content eq '512') {
                $type='ppp';
            } elsif ($content eq '772') {
                $type='loopback';
            }
            $facts->{network_interfaces}{$interface}{type}=$type if ($type);

            # get if the interface is bridged.
            $file="${if_path}/bridge";
            if ($sys->exists($file)) {
                $facts->{network_interfaces}{$interface}{type}='bridge';

                ($rc,$output)=$sys->cmd("_cmd_ls ${if_path}/brif/ 2>/dev/null");
                $facts->{network_interfaces}{$interface}{interfaces}=[ split(/\s+/,$output) ] if (!$rc && $output);

                $file="${if_path}/bridge/bridge_id";
                $content=$sys->readfile($file);
                chomp $content;
                $facts->{network_interfaces}{$interface}{id}=$content if ($content);
            }

            # get if the interface is bonded.
            $file="${if_path}/bonding";
            if ($sys->exists($file)) {
                $facts->{network_interfaces}{$interface}{type}='bonding';

                ($rc,$output)=$sys->cmd("_cmd_ls ${if_path}/bonding/slaves 2>/dev/null");
                $facts->{network_interfaces}{$interface}{interfaces}=[ split(/\s+/,$output) ] if (!$rc && $output);

                $file="${if_path}/bonding/slaves";
                $content=$sys->readfile($file);
                chomp $content;
                $facts->{network_interfaces}{$interface}{slaves}=[ split(/\s+/,$content) ] if ($content);
            }
        }
    }

    return $facts;
}

package Module::facter::network::SunOS;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts);

    $facts={};

    return $facts;
}

package Module::facter::network::AIX;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts);

    $facts={};

    return $facts;
}

package Module::facter::network::HPUX;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts);

    $facts={};

    return $facts;
}

1;
