#
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC. All rights reserved.
# 
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF VERITAS
# TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE
# PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
# 
# The Licensed Software  and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software  -  Restricted Rights"
# and DFARS 227.7202,  Rights  in  "Commercial Computer Software or  Commercial
# Computer Software Documentation" as applicable, and any successor regulations,
# whether  delivered  by  Veritas  as  on premises or hosted services.  Any use,
# modification, reproduction release, performance, display or disclosure of the
# Licensed Software and Documentation by the U.S. Government shall be solely in
# accordance with the terms of this Agreement. $
#

package Module::facter::operatingsystem;

# Get the following facts for operating system
#    platform
#    architecture
#    distribution
#    distribution_version
#    distribution_major_version
#    kernel_version
#    hostname
#    nodename
#    domainname
#    fqdn
#    fips
#    selinux_mode
#    system_manufacturer
#    system_serial_number
#    system_uuid
#    virtualization_type
#    virtualization_role


package Module::facter::operatingsystem::Linux;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts,$file,$content,$rc,$output,$key);

    $facts={};

    $facts->{platform} = $sys->cmd("_cmd_uname -s 2>/dev/null");
    $facts->{architecture} = $sys->cmd("_cmd_uname -m 2>/dev/null");
    $facts->{nodename} = $sys->cmd("_cmd_uname -n 2>/dev/null");
    $facts->{hostname} = $sys->cmd("_cmd_hostname 2>/dev/null");

    $facts->{kernel_version} = $sys->cmd("_cmd_uname -r 2>/dev/null");

    if ($sys->exists('/etc/oracle-release')) {
        $facts->{distribution} = 'OracleLinux';

        $content=$sys->readfile('/etc/oracle-release');
        chomp $content;
        if ($content=~/\s+(\d+)\.(\d+)/) {
            $facts->{distribution_major_version} = $1;
            $facts->{distribution_version} = "$1.$2";
        }
    } elsif ($sys->exists('/etc/redhat-release')) {
        $facts->{distribution} = 'RedHat';

        $content=$sys->readfile('/etc/redhat-release');
        chomp $content;
        if ($content=~/\s+(\d+)\.(\d+)/) {
            $facts->{distribution_major_version} = $1;
            $facts->{distribution_version} = "$1.$2";
        }
    } elsif ($sys->exists('/etc/SuSE-release')) {
        $facts->{distribution} = 'SuSE';

        $content=$sys->readfile('/etc/SuSE-release');
        chomp $content;

        my $version='';
        my $patchlevel='';
        if ($content=~/VERSION\s*=\s*(\d+)/) {
            $version = $1;
            $facts->{distribution_major_version} = $1;
        }
        if ($content=~/PATCHLEVEL\s*=\s*(\d+)/) {
            $patchlevel = $1;
        }

        if ($version && $patchlevel) {
            $facts->{distribution_version} = 'SLES'.$version.'SP'.$patchlevel;
        }
    } elsif ($sys->exists('/etc/os-release')) {
        #$facts->{distribution} = 'SuSE';
    }

    # get kernel_bits
    ($rc,$output)=$sys->cmd("_cmd_getconf LONG_BIT 2>/dev/null");
    if (!$rc && $output) {
        $facts->{kernel_bits}=$output;
    }

    # get BIOS and DMI information
    if ($sys->exists('/sys/devices/virtual/dmi/id/product_name')) {
        # DMI SPEC -- http://www.dmtf.org/sites/default/files/standards/documents/DSP0134_2.7.0.pdf
        my %dmi_files = (
            'bios_release_date' => '/sys/devices/virtual/dmi/id/bios_date',
            'bios_version' => '/sys/devices/virtual/dmi/id/bios_version',
            'bios_vendor' => '/sys/devices/virtual/dmi/id/bios_vendor',
            #'form_factor' => '/sys/devices/virtual/dmi/id/chassis_type',
            'system_product_name' => '/sys/devices/virtual/dmi/id/product_name',
            'system_serial_number' => '/sys/devices/virtual/dmi/id/product_serial',
            'system_uuid' => '/sys/devices/virtual/dmi/id/product_uuid',
            'system_version' => '/sys/devices/virtual/dmi/id/product_version',
            'system_manufacturer' => '/sys/devices/virtual/dmi/id/sys_vendor',
        );

        for my $key (keys %dmi_files) {
            $file = $dmi_files{$key};
            $content = $sys->readfile($file);
            chomp $content;
            if ($content) {
                $facts->{$key} = $content;
            } else {
                $facts->{$key} = 'NA';
            }
        }
    } else {
        # Fall back to using dmidecode, if available
        my %dmi_keys = (
            'bios_release_date' => 'bios-release-date',
            'bios_version' => 'bios-version',
            'bios_vendor' => 'bios-vendor',
            #'form_factor' => 'chassis-type',
            'system_product_name' => 'system-product-name',
            'system_serial_number' => 'system-serial-number',
            'system_uuid' => 'system-uuid',
            'system_version' => 'system-version',
            'system_manufacturer' => 'system-manufacturer'
        );
        for my $key (keys %dmi_keys) {
            my $type=$dmi_keys{$key};
            ($rc, $output)=$sys->cmd("_cmd_dmidecode -s $type 2>/dev/null");
            if (!$rc && $output) {
                $facts->{$key}=$output;
            } else {
                $facts->{$key}='NA';
            }
        }
    }

    # get virtualization information
    # For more information, check: http://people.redhat.com/~rjones/virt-what/
    my $system_product_name=$facts->{system_product_name};
    if ($system_product_name =~ /VMware/mxi) {
        $facts->{virtualization_type} = 'VMware';
        $facts->{virtualization_role} = 'guest';
    } elsif ($system_product_name =~ /VirtualBox/mxi) {
        $facts->{virtualization_type} = 'VirtualBox';
        $facts->{virtualization_role} = 'guest';
    } elsif ($system_product_name =~ /RHEV/mxi) {
        $facts->{virtualization_type} = 'RHEV';
        $facts->{virtualization_role} = 'guest';
    } elsif ($system_product_name =~ /(KVM|Bochs)/mxi) {
        $facts->{virtualization_type} = 'KVM';
        $facts->{virtualization_role} = 'guest';
    }

    # get selinux
    $content=$sys->readfile("/etc/selinux/config");
    chomp $content;
    if ($content && $content=~/^SELINUX=\s*(\S+)/mx) {
        $facts->{selinux_mode}=$1;
    }

    # get fips
    $file='/proc/sys/crypto/fips_enabled';
    $content=$sys->readfile($file);
    chomp $content;
    if ($content && $content eq '1') {
        $facts->{fips} = $1;
    }

    return $facts;
}

package Module::facter::operatingsystem::SunOS;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts);

    $facts={};
    $facts->{platform} = $sys->cmd("_cmd_uname -s 2>/dev/null");
    $facts->{architecture} = $sys->cmd("_cmd_uname -m 2>/dev/null");
    $facts->{nodename} = $sys->cmd("_cmd_uname -n 2>/dev/null");
    $facts->{hostname} = $sys->cmd("_cmd_hostname 2>/dev/null");

    $facts->{distribution}='SunOS';

    return $facts;
}

package Module::facter::operatingsystem::AIX;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts);

    $facts={};
    $facts->{platform} = $sys->cmd("_cmd_uname -s 2>/dev/null");
    $facts->{architecture} = $sys->cmd("_cmd_uname -m 2>/dev/null");
    $facts->{nodename} = $sys->cmd("_cmd_uname -n 2>/dev/null");
    $facts->{hostname} = $sys->cmd("_cmd_hostname 2>/dev/null");

    $facts->{distribution}='AIX';

    return $facts;
}

package Module::facter::operatingsystem::HPUX;

sub execute_sys {
    my ($self,$sys) = @_;
    my ($facts);

    $facts={};
    $facts->{platform} = $sys->cmd("_cmd_uname -s 2>/dev/null");
    $facts->{architecture} = $sys->cmd("_cmd_uname -m 2>/dev/null");
    $facts->{nodename} = $sys->cmd("_cmd_uname -n 2>/dev/null");
    $facts->{hostname} = $sys->cmd("_cmd_hostname 2>/dev/null");

    $facts->{distribution}='HPUX';

    return $facts;
}

1;
