package Module::system::lineinfile;

# description:
#     - Ensure a line is in a file, or replace an existing line use regular expression
# options:
#     dest:
#         required: true
#         aliases: [ "name", "destfile" ]
#         description:
#              - Name of the file to modify
#     state:
#         required: false
#         default: "present"
#         choices: [ present, absent ]
#         description:
#             - Whether the line should be there or not
#     regexp:
#         required: false
#         description:
#             - The regular expression to look for a line in the file
#               For state=present, only the last line found will be replaced.
#               For state=absent, all the lines found will be removed.
#     backrefs:
#         required: false
#         default: "no"
#         choices: [ "yes", "no" ]
#         description:
#             - Used with state=present. If set, line can contain backreferences
#               (both positional and named) that will get populated if the 'regexp'
#               matches.
#     insertafter:
#         required: false
#         description:
#             - Used with state=present. If specified, the line will be inserted
#               after the last match of the specified regular expression if the 
#               line is not matched with 'regexp'
#     insertbefore:
#         required: false
#         description:
#             - Used with state=present. If specified, the line will be inserted
#               before the last match of the specified regular expression if the
#               line is not matched with 'regexp'
#         choices: [ "BOF", "regex" ]
#     create:
#         required: false
#         choices: [ "yes", "no" ]
#         default: "no"
#         description:
#             - Used with state=present. If specified, the file will be created
#               if it does not already exist. By default it will fail if the file
#               is missing.
#     backup:
#         required: false
#         default: "no"
#         choices: [ "yes", "no" ]
#         description:
#             - Create a backup file including the timestamp information.

sub execute_sys {
    my ($self,$sys,$attrs) = @_;
    my ($return,$destfile,$state);

    $return={ 'changed' => 0 };

    return $return if (!defined $attrs);

	$attrs=eval($attrs) if ref($attrs) eq '';

    $destfile=$attrs->{dest} || $attrs->{name} || $attrs->{destfile} || '';
    return $return if (!defined $destfile);

    $state=$attrs->{state} || 'present';
    $state=lc($state);
    if ($state eq 'present') {
        $return=$self->lineinfile_present_sys($sys,$destfile,$attrs);    
    } elsif ($state eq 'absent') {
        $return=$self->lineinfile_absent_sys($sys,$destfile,$attrs);    
    }
    return $return;
}

sub lineinfile_present_sys {
    my ($self,$sys,$destfile,$attrs) = @_;
    my ($regexp,$line,$insertafter,$insertbefore,$create,$backup,$backrefs,$backrefs1);
    my ($lines,@lines,$lineno,$index,$matched,$matched_insertafter,$matched_insertbefore);
    my ($return,$changed,$results,$msg,$rc);

    $changed=0;

    if ($sys->exists($destfile)) {
        # File exist
        $lines=$sys->readfile($destfile);
        @lines=split(/\n/,$lines);
        if (scalar @lines) {
            $backrefs1=EDRu::normalize_value($attrs->{backrefs});

            # Can use the following lines to specify multiple changes
            # regexp1=^line1
            # line1='line1'
            # insertafter1=EOF
            # regexp2=^line2
            # line2='line2'
            # insertbefore2=BOF
            for my $key (%{$attrs}) {
                next unless ($key=~/^regexp(\d*)$/);

                $index=$1||'';
                $regexp=$attrs->{"regexp${index}"};
                $line=$attrs->{"line${index}"};

                next if (!$regexp || !$line);

                $insertafter=$attrs->{"insertafter${index}"} || $attrs->{"insertafter"};
                $insertbefore=$attrs->{"insertbefore${index}"} || $attrs->{"insertbefore"};
                $backrefs=EDRu::normalize_value($attrs->{"backrefs${index}"}) || $backrefs1;

                # Find matched lines
                $matched=undef;
                $matched_insertafter=undef;
                $matched_insertbefore=undef;

                $lineno=0;
                for my $line1 (@lines) {
                    if ($regexp && $line1=~/$regexp/) {
                        # If find the line, replace the matched line
                        $matched=$lineno;
                        if ($backrefs) {
                            my $first=$1;
                            my $second=$2;
                            my $third=$3;
                            my $fourth=$4;
                            my $fifth=$5;
                            my $line2=$line;
                            $line2=~s/\\5/$fifth/g if (defined $fifth);
                            $line2=~s/\\4/$fourth/g if (defined $fourth);
                            $line2=~s/\\3/$third/g if (defined $third);
                            $line2=~s/\\2/$second/g if (defined $second);
                            $line2=~s/\\1/$first/g if (defined $first);
                            $lines[$matched]=$line2;
                            $changed=1;
                        } else {
                            if ($line1 ne $line) {
                                $lines[$matched]=$line;
                                $changed=1;
                            }
                        }
                    }
                    if ($insertafter && 
                        ($insertafter ne 'BOF') &&
                        ($insertafter ne 'EOF') &&
                        ($line1=~/$insertafter/)) {
                        $matched_insertafter=$lineno;
                    }
                    if ($insertbefore && 
                        ($insertbefore ne 'BOF') &&
                        ($line1=~/$insertbefore/)) {
                        $matched_insertbefore=$lineno;
                    }
                    $lineno++;
                }

                if (!defined $matched) {
                    # If not find the line, insert the line 
                    if ($insertafter) {
                        if ($insertafter eq 'EOF') {
                            push @lines, $line;
                            $changed=1;
                        } elsif ($insertafter eq 'BOF') {
                            unshift @lines, $line;
                            $changed=1;
                        } elsif ($matched_insertafter) {
                            $changed=1;
                            splice @lines, $matched_insertafter+1, 0, $line;
                        }
                    } elsif ($insertbefore) {
                        if ($insertbefore eq 'BOF') {
                            $changed=1;
                            unshift @lines, $line;
                        } elsif ($matched_insertbefore) {
                            $changed=1;
                            splice @lines, $matched_insertbefore, 0, $line;
                        }
                    }
                }
            }

            if ($changed == 1) {
                # write the lines back to the file
                $backup=$attrs->{backup};
                if ($backup) {
                    my $ext='bak'; 
                    $sys->copyfile($destfile,"$destfile\.$ext");
                }
                $lines = join "\n", @lines;
                $lines .= "\n";
                $sys->writefile($lines, $destfile);
            }
        }
    } else {
        # File not exist
        $create=EDRu::normalize_value($attrs->{create});
        $backrefs=EDRu::normalize_value($attrs->{backrefs});
        if ($create == 1 && $backrefs != 1 && $line) {
            $changed=1;
            $sys->writefile($line,$destfile);
        } else {
            $rc=127;
            $msg="Destination $destfile does not exist!";
        }
    }

    $return->{changed} = $changed;
    $return->{rc} = $rc if (defined $rc);
    $return->{msg} = $msg if (defined $msg);
    $return->{results} = $results if (defined $results);
    return $return;
}

sub lineinfile_absent_sys {
    my ($self,$sys,$destfile,$attrs) = @_;
    my ($lines,@lines,$index,$regexp,$backup);
    my ($return,$changed,$results,$msg,$rc);

    $changed=0;

    if ($sys->exists($destfile)) {
        # File exist
        $lines=$sys->readfile($destfile);
        @lines=split(/\n/,$lines);
        if (scalar @lines) {
            # Can use the following lines to specify multiple changes
            # regexp1=^line1
            # insertafter1=EOF
            # regexp2=^line2
            # line2='line2'
            for my $key (%{$attrs}) {
                next unless ($key=~/^regexp(\d*)$/);

                $index=$1||'';
                $regexp=$attrs->{"regexp${index}"};

                next if (!$regexp);

                # Find matched lines
                my @newlines=grep { ! /$regexp/ } @lines;
                if ((scalar @newlines) != (scalar @lines)) {
                    $changed=1;
                    @lines=@newlines;
                }
            }

            if ($changed == 1) {
                # write the lines back to the file
                $backup=$attrs->{backup};
                if ($backup) {
                    my $ext='bak'; 
                    $sys->copyfile($destfile,"$destfile\.$ext");
                }
                $lines = join "\n", @lines;
                $lines .= "\n";
                $sys->writefile($lines, $destfile);
            }
        }
    }

    $return->{changed} = $changed;
    $return->{rc} = $rc if (defined $rc);
    $return->{msg} = $msg if (defined $msg);
    $return->{results} = $results if (defined $results);
    return $return;
}

1;
