package Module::system::user;

# requirements: [ useradd, userdel, usermod ]
# description:
#     - Manage user accounts and user attributes.
# options:
#     name:
#         required: true
#         aliases: [ "user" ]
#         description:
#              - Name of the user to create, remove or modify.
#     state:
#         required: false
#         default: "present"
#         choices: [ present, absent ]
#         description:
#             - Whether the account should exist or not, taking action if the state is different from what is stated.
#     comment:
#         required: false
#         description:
#             - Optionally sets the description (aka I(GECOS)) of user account.
#     uid:
#         required: false
#         description:
#             - Optionally sets the I(UID) of the user.
#     group:
#         required: false
#         description:
#             - Optionally sets the user's primary group (takes a group name).
#     shell:
#         required: false
#         description:
#             - Optionally set the user's shell.
#     home:
#         required: false
#         description:
#             - Optionally set the user's home directory.

sub execute_sys {
    my ($self,$sys,$attrs) = @_;
    my ($return,$plat,$sub,$name,$state);

    $return={ 'changed' => 0 };

    # get distribution
    $plat=$sys->{plat};
    return $return if (!$plat);

    return $return if (!defined $attrs);

	$attrs=eval($attrs) if ref($attrs) eq '';

    $name=$attrs->{name};
    return $return if (!$name);

    $state=$attrs->{state} || 'present';

    if ($state eq 'present') {
        $sub='create_user';
    } elsif ($state eq 'absent') {
        $sub='delete_user';
    }

    $plat=lc($plat);
    $sub="${sub}_${plat}";
    if ($self->can($sub)) {
        $self->$sub($sys,$attrs);
        $return->{changed} = 1;
    }

    return $return;
}

# Linux
sub create_user_linux {
    my ($self,$sys,$attrs) = @_;
    my ($name,$uid,$home,$shell,$group,$comment,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd='_cmd_useradd ';

    $uid=$attrs->{uid};
    $cmd.=" -u $uid" if ($uid);

    $home=$attrs->{homedir};
    $cmd.=" -d $home" if ($home);

    $shell=$attrs->{shell};
    $cmd.=" -s $shell" if ($shell);

    $group=$attrs->{group};
    $cmd.=" -g $group" if ($group);

    $comment=$attrs->{comment};
    $cmd.=" -c $comment" if ($comment);

    $cmd.=" $name";
   
    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}

sub delete_user_linux {
    my ($self,$sys,$attrs) = @_;
    my ($name,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd="_cmd_userdel -f $name";

    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}

# Solaris
sub create_user_sunos {
    my ($self,$sys,$attrs) = @_;
    my ($name,$uid,$home,$shell,$group,$comment,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd='_cmd_useradd ';

    $uid=$attrs->{uid};
    $cmd.=" -u $uid" if ($uid);

    $home=$attrs->{homedir};
    $cmd.=" -d $home" if ($home);

    $shell=$attrs->{shell};
    $cmd.=" -s $shell" if ($shell);

    $group=$attrs->{group};
    $cmd.=" -g $group" if ($group);

    $comment=$attrs->{comment};
    $cmd.=" -c $comment" if ($comment);

    $cmd.=" $name";
   
    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}

sub delete_user_sunos {
    my ($self,$sys,$attrs) = @_;
    my ($name,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd="_cmd_userdel -f $name";

    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}


# AIX
sub create_user_aix {
    my ($self,$sys,$attrs) = @_;
    my ($name,$uid,$home,$shell,$group,$comment,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd='_cmd_useradd ';

    $uid=$attrs->{uid};
    $cmd.=" -u $uid" if ($uid);

    $home=$attrs->{homedir};
    $cmd.=" -d $home" if ($home);

    $shell=$attrs->{shell};
    $cmd.=" -s $shell" if ($shell);

    $group=$attrs->{group};
    $cmd.=" -g $group" if ($group);

    $comment=$attrs->{comment};
    $cmd.=" -c $comment" if ($comment);

    $cmd.=" $name";
   
    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}

sub delete_user_aix {
    my ($self,$sys,$attrs) = @_;
    my ($name,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd="_cmd_userdel -f $name";

    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}


# HPUX
sub create_user_hpux {
    my ($self,$sys,$attrs) = @_;
    my ($name,$uid,$home,$shell,$group,$comment,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd='_cmd_useradd ';

    $uid=$attrs->{uid};
    $cmd.=" -u $uid" if ($uid);

    $home=$attrs->{homedir};
    $cmd.=" -d $home" if ($home);

    $shell=$attrs->{shell};
    $cmd.=" -s $shell" if ($shell);

    $group=$attrs->{group};
    $cmd.=" -g $group" if ($group);

    $comment=$attrs->{comment};
    $cmd.=" -c $comment" if ($comment);

    $cmd.=" $name";
   
    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}

sub delete_user_hpux {
    my ($self,$sys,$attrs) = @_;
    my ($name,$cmd,$rc,$output);

    $name=$attrs->{name};
    return 0 if (!$name);

    $cmd="_cmd_userdel -f $name";

    ($rc, $output)=$sys->cmd("$cmd 2>/dev/null");

    return $rc ? 0 : 1;
}

1;
