package Module::veritas::llt;
# description:
#   - LLT configuration
#   Recommend requirement: >=2 private LLT link + 1 low priority link
#   Minimum requirement:
#       - private LLT link + 1 low priority link (in which case Fencing must be configured)
#       - 2 private LLT links
#
# options:
#   - state:
#       required: true
#       choices: [present, absent]
#       description:
#           enable/disable LLT
#       choice -> present:
#           - method:
#               required: true
#               choices: [ethernet, udp, rdma]
#               description:
#                   1. LLT over ethernet is the fastest way to for communication,
#                   but it requires direct connection between 2 NICs, which is impossible in some scenarios
#                   2. LLT over UDP means the LLT packets are routable and thus able to span longer distance,
#                   hence it requires IP address, port and netmask for routing. 
#
#                   <explain difference among ethernet/udp/rdma>  
#               choice --> udp:
#                   udp-link:
#						- system: [ hostname ]
#						  heartbeat_link:
#							- nic:  [ NIC name ]
#							  ip:   [ ip address ]
#							  port: [ UDP port ]
#						  low_priority_link:
#						    - nic:  [ NIC name ]
#						      ip:   [ ip address ]
#						      port: [ UDP port ]
#
#           - vcs_clusterid:
#               required: true
#               choices: [ Cluster ID ]
#           - llt_hosts:
#               required: true
#               choices: [ (Cluster systems' hostname array) ]
#           - private_link:
#               required: true
#               choices: [ (NIC name array) ]
#               choice --> only one NIC:
#                   - low_priority_link:
#                       required: true
#                       choices: [ (low priority NIC name array) ]

#   Issues:
#   Only for Linux platform.
#   commands need to be cross-platform, we need to use _cmd_xx to execute
#   define return value for each function so that in the master node, it can read the return value and generate corresponding text
#   result only have 1 changed value, seems not enough.

$Macros::macros->set("VERITAS_LLT_CONF_SUCCESS",0);
$Macros::macros->set("VERITAS_LLT_CONF_FAIL",1);
$Macros::macros->set("VERITAS_LLT_UNCONF_SUCCESS",2);
$Macros::macros->set("VERITAS_LLT_UNCONF_FAIL",3);
$Macros::macros->set("VERITAS_LLT_NIC_NOT_EXIST",5);
$Macros::macros->set("VERITAS_LLT_WRITE_CONF_FAIL",7);
$Macros::macros->set("VERITAS_LLT_NO_ENABLE_FILE",9);
$Macros::macros->set("VERITAS_LLT_DELET_FILE_FAIL",11);
$Macros::macros->set("VERITAS_LLT_UDP_IP_NOT_EXIST",13);

sub execute_sys {
    my ($self,$sys,$attr) = @_;
    my ($rc,$return,$llttab,$llthosts);
    my ($hostname,$nichash,$result,$n,$string);

    $return={ 'changed' => 0, 'rc' => 0 };
    return $return if (!defined $attr);
	$attr=eval($attr) if ref($attr) eq '';
    for my $key (keys%{$attr}) {
        if (EDRu::inarr($key,("private_link", "llt_hosts", "low_priority_link"))) {
            my @value=split(',',$attr->{$key});
            $attr->{$key}=\@value;
        }
    }

    #todo: $attribute type check needs to be performed here
    if ($attr->{state} eq 'present') {
        return $return if (!$attr->{vcs_clusterid} || !$attr->{llt_hosts});

		$hostname = EDRu::nofqdn($sys->{hostname});
		$llttab="set-node $hostname\nset-cluster $attr->{vcs_clusterid}\n";

        if ($attr->{method} eq 'ethernet') {
			return $return if (!$attr->{private_link} || # No private LLT link
					(scalar(@{$attr->{private_link}})==1) && (!$attr->{low_priority_link})); # 1 private LLT link, no priority link
            for my $nic (@{$attr->{private_link}}) {
				$rc = check_nic_is_availaible($sys,$nic);
				if ( $rc ){
					$return->{rc} = $macros->{VERITAS_LLT_NIC_NOT_EXIST};
					return $return;
				}
                $llttab .= convert_nic2lltlink_sys($sys,$nic);
            }
            if ($attr->{low_priority_link}) {
                for my $nic (@{$attr->{low_priority_link}}) {
					$rc = check_nic_is_availaible($sys,$nic);
					if ( $rc ){
						$return->{rc} = $macros->{VERITAS_LLT_NIC_NOT_EXIST};
						return $return;
					}
                    $llttab .= convert_nic2lltlink_sys($sys,$nic,'lowpri');
                }
            }
            chomp($llttab);

        } elsif ($attr->{method} eq 'udp') {
			$string = get_udp_llttab( $sys, $attr );
			if ( $string ){
				$llttab .= $string;
			} else {
				$return->{rc} = $macros->{VERITAS_LLT_UDP_IP_NOT_EXIST};
				return $return;
			}

        } elsif ($attr->{method} eq 'rdma') {
        }

		$n=0;
		for my $host (@{$attr->{llt_hosts}}) {
			$llthosts.="$n $host\n";
			$n++;
		}
		chomp($llthosts);

		($rc,$result) = $sys->cmd("_cmd_echo '$llttab' > /etc/llttab 2>/dev/null; _cmd_echo '$llthosts' > /etc/llthosts 2>/dev/null");
        if ($rc) {
            $sys->log("Failed to write the LLT configuration files", 'error');
            $return->{rc} = $macros->{VERITAS_LLT_WRITE_CONF_FAIL};
        } else {
            $return->{changed}=1;
        }
        $rc = $self->enable_llt($sys);
        $return->{changed} ||= $rc->{changed};
        $return->{rc} ||= $rc->{rc};

        if ($return->{rc}) {
            $sys->log("Failed to configure LLT.",'error');
            $return->{rc} = $macros->{VERITAS_LLT_CONF_FAIL};
        } else {
            $return->{rc} = $macros->{VERITAS_LLT_CONF_SUCCESS};		
        }   
    } elsif ($attr->{state} eq 'absent') {
        ($rc,$result)=$sys->cmd("_cmd_rmr /etc/llt* 2>/dev/null");
        if (!$rc) {
            $return->{changed}=1;
        } else {
            $sys->log("Failed to delete the LLT configuration files");
            $return->{rc} = $macros->{VERITAS_LLT_DELET_FILE_FAIL};			
        }
        $rc = $self->disable_llt($sys);
        $return->{changed} ||= $rc->{changed};
        $return->{rc} ||= $rc->{rc};

        if ($return->{rc}) {
            $sys->log("Failed to un-configure LLT.",'error');
            $return->{rc} = $macros->{VERITAS_LLT_UNCONF_FAIL};
        } else {
            $return->{rc} = $macros->{VERITAS_LLT_UNCONF_SUCCESS};		
        }
    }
    return $return;
}

#only for Linux
sub get_mac_by_nic_sys {
    my ($sys,$nic) = @_;
    my ($output);
    if ($sys->{plat} eq 'Linux') {
        $output = $sys->cmd("_cmd_ip link show $nic 2>/dev/null | _cmd_awk '/^ *link/ { print \$2 }'");
    } else {
        $output = "NOT SUPPORT";
    }
    return $output;
}

sub is_vlan_nic_sys {
    my ($sys,$nic) = @_;
    return $sys->exists("/proc/net/vlan/$nic");
}

sub is_bonded_nic_sys {
    my ($sys,$nic)=@_;
    my ($bnics);
    if ($sys->{bondednics}){
        return EDRu::inarr($nic,@{$sys->{bondednics}});
    }
    ($bnics,undef)=bondednics_sys($sys);
    return EDRu::inarr($nic,@$bnics);
}

#only for Linux
sub bondednics_sys {
    my ($sys)=@_;
    my ($files,@snics,$file_pattern,@slavenics,$nic,$slave,@bondednics);
    if ($sys->{plat} eq 'Linux') {
        #$file_pattern="/etc/sysconfig/network-script/ifcfg-bond*";
        $file_pattern='/proc/net/bonding/';
        $files=$sys->cmd("_cmd_ls $file_pattern 2>/dev/null");
        for my $file (split(/\n/, $files)){
            $slave=$sys->cmd("_cmd_grep 'Slave Interface' $file_pattern$file 2>/dev/null | _cmd_awk '{print \$3}'");
            @snics=split(/\n/,$slave);
            push(@slavenics,@snics) if($#snics>=0);
            (undef,undef,undef,undef,$nic)=split(/\//m,$file);
            push(@bondednics,$file);
        }
        $sys->{'bondednics'}=[@bondednics];
    }
    return (\@bondednics,\@slavenics);
}

#only for Linux
sub convert_nic2lltlink_sys {
    my ($sys,$nic,$lowpri) = @_;
    my ($rc,$output,@ifc);
    
    if ($lowpri) {
        $lowpri = '-lowpri';
    } else {
        $lowpri = '';
    }
    
    if (is_vlan_nic_sys($sys, $nic) || is_bonded_nic_sys($sys, $nic)) {
        # VLAN interface. Cant use MAC here.
        return "link$lowpri $nic $nic - ether - -\n";
    }
    ($rc,$output) = check_nic_is_availaible($sys,$nic);
    
	@ifc = split (/\s+/m,$output);
    my $idx = EDRu::arrpos('link/ether', @ifc);
    if ($idx == -1) {
        # NIC without a MC?
        return '';
    }
    # The next index will be the MAC
    return "link$lowpri $nic eth-$ifc[$idx+1] - ether - -\n";
}

#only for Linux
sub nic_bcast_sys {
    my ($sys,$nic,$ip)=@_;
    my ($out,$bcast);
    return if(EDRu::ip_is_ipv6($ip));
    $out=$sys->cmd("_cmd_ip -o addr show dev $nic 2>/dev/null | _cmd_grep '$ip'");
    if($out=~/brd (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/){
       $bcast=$1;
    }
    return $bcast;
}

sub convert_nic2udplink_sys {
    my ($sys,$nic,$port,$address,$bcast,$lowpri) = @_;
    my ($dev,$linktype,$bcastaddr);
	$linktype = 'udp';
    if($sys->{plat} eq 'AIX') {
        $dev = '/dev/xti/udp';
    } elsif ($sys->{plat} =~ /SunOS|HPUX/m) {
        $dev = '/dev/udp';
    } else {
        $dev = 'udp';
    }
    if(EDRu::ip_is_ipv6($address)) {
        $dev=$dev."6";
        $linktype = 'udp6';
    } elsif ($bcast) {
		#$bcastaddr=nic_bcast_sys($sys,$nic,$address);
		$bcastaddr = '-';
    }

	return "link-lowpri $nic $dev - $linktype $port - $address $bcastaddr\n"  if ($lowpri);

    return "link $nic $dev - $linktype $port - $address $bcastaddr\n";
}

sub enable_llt {
    my ($self, $sys) = @_;
    my ($return, $dir, $rc, $conf, $output);
    $return = {'changed' => 0, 'rc' => 0};
    $dir = "/etc/sysconfig";
    ($rc, $conf) = $sys->cmd("_cmd_cat $dir/llt");

    if ($rc){
        $sys->log("Cannot find the file $dir/llt", 'error');
        $return->{rc} = $macros->{VERITAS_LLT_NO_ENABLE_FILE};
        return $return;
    }
    $sys->cmd("_cmd_rmr $dir/llt.prev");
    $sys->cmd("_cmd_cp $dir/llt $dir/llt.prev");
    $sys->cmd("_cmd_sed -i 's/LLT_START=0/LLT_START=1/g' $dir/llt");
    $sys->cmd("_cmd_sed -i 's/LLT_STOP=0/LLT_STOP=1/g' $dir/llt");
    ($rc, $output) = $sys->cmd("_cmd_diff $dir/llt $dir/llt.prev");
    $return->{changed} = $rc;

    return $return;
}

sub disable_llt {
    my ($self, $sys) = @_;
    my ($return, $dir, $rc, $conf, $output);
    $return = {'changed' => 0, 'rc' => 0};
    $dir = "/etc/sysconfig";
    ($rc, $conf) = $sys->cmd("_cmd_cat $dir/llt");

    if ($rc){
        $sys->log("Cannot find the file $dir/llt", 'error');
        $return->{rc} = $macros->{VERITAS_LLT_NO_ENABLE_FILE};
        return $return;
    }
    $sys->cmd("_cmd_rmr $dir/llt.prev");
    $sys->cmd("_cmd_cp $dir/llt $dir/llt.prev");
    $sys->cmd("_cmd_sed -i 's/LLT_START=1/LLT_START=0/g' $dir/llt");
    $sys->cmd("_cmd_sed -i 's/LLT_STOP=1/LLT_STOP=0/g' $dir/llt");
    ($rc, $output) = $sys->cmd("_cmd_diff $dir/llt $dir/llt.prev");
    $return->{changed} = $rc;

    return $return;
}

sub check_nic_is_availaible {
	my ( $sys, $nic ) = @_;
	my ( $rc, $output );
	if ( $sys->{plat} eq 'Linux' ){
		( $rc, $output ) = $sys->cmd("_cmd_ip -o link show dev '$nic'");
	} else {
		return 0;
	}

	chomp $output;	
	$sys->log("NIC $nic does not exist. LLT may not come up.") if $rc;

	return ( $rc, $output ) if wantarray;
	return $rc;
}

sub check_ip_is_exist {
	my ($sys,$nic,$ip)=@_;
	my ($rc,$out);
	if ( $sys->{plat} eq 'Linux' ){
		($rc,$out) = $sys->cmd("_cmd_ip -o addr show dev $nic 2>/dev/null | _cmd_grep '$ip'");
	} else {
		return 0;
	}
	chomp $out;

	$sys->log("IP $ip does not exist on $nic.") if $rc;

	return ($rc, $out) if wantarray;
	return $rc;
}

sub get_udp_llttab {
	my ( $sys, $attrs, $hostname ) = @_;
	my ( $content, $local, $others, $nid, $string);
	my ( $rc, $nic, $address );
	$nid = 0;

	for my $host ( @{$attrs->{udp_link}} ){
		if( $sys->{hostname} =~ $host->{system} ){
			for my $hb ( @{$host->{heartbeat_link}} ){
				$rc = check_ip_is_exist($sys,$hb->{nic},$hb->{ip});
				return '' if $rc;
			    $local .= convert_nic2udplink_sys($sys,$hb->{nic},$hb->{port},$hb->{ip},1,0);
			}
			for my $lowpri ( @{$host->{low_priority_link}} ){
				$rc = check_ip_is_exist($sys,$lowpri->{nic},$lowpri->{ip});
				return '' if $rc;
				$local .= convert_nic2udplink_sys($sys,$lowpri->{nic},$lowpri->{port},$lowpri->{ip},1,1);
			}
		} else {
			for my $hb ( @{$host->{heartbeat_link}} ){
				$nic = $hb->{nic};
				$address = $hb->{ip};
				$others .= "set-addr $nid $nic $address\n";
			}
			for my $lowpri ( @{$host->{low_priority_link}} ){
				$nic = $lowpri->{nic};
				$address = $lowpri->{ip};
				$others .= "set-addr $nid $nic $address\n";
			}
		}
		$nid++;
	}

	$content = $local . $others;
	$content.="set-bcasthb 0\n";
	$content.="set-arp 0\n";
	chomp $content;
	return $content;
}

1;
