#
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC. All rights reserved.
# 
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF VERITAS
# TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE
# PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
# 
# The Licensed Software  and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software  -  Restricted Rights"
# and DFARS 227.7202,  Rights  in  "Commercial Computer Software or  Commercial
# Computer Software Documentation" as applicable, and any successor regulations,
# whether  delivered  by  Veritas  as  on premises or hosted services.  Any use,
# modification, reproduction release, performance, display or disclosure of the
# Licensed Software and Documentation by the U.S. Government shall be solely in
# accordance with the terms of this Agreement. $
#
package Sys::Common;

use strict;
use Fcntl qw(:DEFAULT :flock);
use Time::HiRes qw(sleep);
use Cwd qw(abs_path);
use File::Path;
use File::Basename;

sub new {
    my $class = shift;
    my $self  = {};
    bless($self, $class);
    $self->{class} = $class;
	$self->init();
	$self->init_cmds();
    return $self;
}

sub init {
    my $sys = shift;
    $sys->{module} = $ARGV[0] || '';
    $sys->{arg} = $ARGV[1] || '';
    $sys->{tmpdir} = $sys->app_path();
    $sys->{hostname} = `hostname`;
    chomp $sys->{hostname};
    $sys->{logfile} = "$sys->{tmpdir}/$sys->{hostname}.log";
    $sys->{summaryfile} = "$sys->{tmpdir}/$sys->{hostname}.summary";
    $sys->{output} = "$sys->{tmpdir}/$sys->{hostname}.output";
    $sys->create_logfile();
    return;
}

sub aix { return 1 if ($_[0]->{plat} eq 'AIX'); return 0; }
sub aix_vios {return 1 if (($_[0]->{plat} eq 'AIX') && ($_[0]->{ioslevel})); return 0; }

sub hpux { return 1 if ($_[0]->{plat} eq 'HPUX'); return 0; }
sub linux { return 1 if ($_[0]->{plat} eq 'Linux'); return 0; }
sub sunos { return 1 if ($_[0]->{plat} eq 'SunOS'); return 0; }
sub sunos_sol11 {return 1 if (($_[0]->{plat} eq 'SunOS') && ($_[0]->{platvers} eq '5.11')); return 0; }
sub freebsd { return 1 if ($_[0]->{plat} eq 'FreeBSD'); return 0; }
sub macos { return 1 if ($_[0]->{plat} eq 'MacOS'); return 0; }
# modify if padvs change
sub windows { return 1 if ($_[0]->{plat}=~/Win/mx); return 0; }

sub exists {
    my ($sys,$file)=@_;
    return unless ($file);
    return (-e $file) ? 1 : 0;
}

sub is_dir {
    my ($sys,$dir)=@_;
    return unless ($dir);
    return (-d $dir) ? 1 : 0;
}

sub is_executable {
    my ($sys,$file)=@_;
    return unless ($file);
    return (-x $file) ? 1 : 0;
}

# write a string to a file
sub copyfile {
    my($sys,$file_src,$file_dest)= @_;
    if ($file_src && $file_dest) {
        $sys->cmd("_cmd_cp -rp $file_src $file_dest 2>/dev/null");
    }
    return 0;
}

sub cmd {
    my ($sys,$cmd,$env) = @_;
    my ($stdout,$exitcode,$timer,$elapsed,$original_cmd);

    $sys->die("Sys object not passed as first arg to cmd(@_)") if (ref($sys) !~ 'Sys');

    $cmd=$sys->cmdswap($cmd) if ($cmd=~/_cmd_/m);
    $cmd=~s/\/dev\/null/nul/g if ($sys->windows);

    # set timer
    $timer=Timer->new('Sys::command');
    $timer->start();
    $original_cmd=$cmd;

    # Execute the command
    $cmd.=" 2>&1" unless ($cmd=~/2>/);
    ($stdout,$exitcode,$cmd)=$sys->run_local_command($cmd);
    $exitcode||=0;

    # Log the command information, run_local_command() will log the command before running
    #Msg::log("cmd $cmd");
    $sys->log("cmd stdout=\n$stdout") if ($stdout && ($cmd!~/tar -xvf/m));
    $elapsed=$timer->stop();
    $sys->log("cmd exit=$exitcode (duration: $elapsed seconds)");

    if (wantarray) {
        return ($exitcode, $stdout);
    }
    return $stdout;
}

sub cmdswap {
    my ($sys,$cmd) = @_;
    my ($key,$swap);
    return $cmd if ($cmd!~/_cmd_/m);
    $sys->die("Cannot execute $cmd with undefined sys->{class}") if (!$sys->{class});

    while($cmd=~m/\b_cmd_(\w+)\b/sx) {
        $key=$1;
        $swap=$sys->{cmd}{$key};
        if (!$swap) {
            $sys->die("undefined cmd substitution:\n$cmd\nclass=$sys->{class}\n");
        }
        $cmd=~s/\b_cmd_$key\b/$swap/sxg;
    }
    return $cmd;
}

sub run_local_command {
    my ($sys, $cmd)=@_;
    my ($stdout,$exitcode);

    # Log the command before running
    $sys->log("cmd LC_ALL=C LANG=C $cmd");

    # Get stdout
    $stdout=`LC_ALL=C LANG=C $cmd`;
    chomp $stdout;

    # Get exit code
    $exitcode=$?;
    $exitcode >>= 8 if ($exitcode > 0);

    return ($stdout,$exitcode,$cmd);
}

# determine size of package or patch on media
sub filesize {
    my ($sys,$file)=@_;
    # Calculate filesize in KB.
    my $size = (stat $file)[7] > 1023 ? (stat $file)[7]/1024 : 1;
    # assume compression of 3x, so *4 as both
    # tarfile and original are there after gunzip
    $size*=4 if (($size) && ($file=~/gz$/m));
    return $size;
}

# read a file
sub readfile {
    my ($sys,$file,$lock) = @_;
    my ($line,$msg,$rf,$fd);
    if(!open($fd, '<', $file)) {
        $sys->log("readfile cannot open $file");
        #todo: return a code
        return;
    }
    if($lock){
        if(!flock($fd, LOCK_SH)){
            $sys->log("cannot lock $file");
            #todo: return a code
            return;
        }
    }
    $rf ='';
    while ($line=<$fd>) { $rf.=$line; }
    close($fd);
    return $rf;
}

sub readfile_with_lock {
    my ($sys,$file) = @_;
    return readfile($file,1);
}

# write a string to a file
sub writefile {
    my($sys,$write,$file)= @_;
    # need two checks here:
    # is the directory created
    # is the filename a directory, if so die because the write will fail
    # should check this earlier, if possible
    my $msg;
    if(!sysopen(WF, "$file", O_RDWR | O_CREAT)){
        $sys->log("writefile cannot open $file");
        #todo: return a code
        return;
    }
    flock(WF, LOCK_EX);
    # Now we have acquired the lock, it's safe for I/O
    seek(WF, 0, 0);
    truncate(WF, 0);
    print WF $write;
    close(WF);
    return 0;
}

# append a newlined string to a file
sub appendfile {
    my($sys,$msg,$file)= @_;
    my $fd;
    if(!open($fd, '>>', $file)){
        #todo: return a code
        return;
    }
    print $fd "$msg\n";
    close($fd);
    return 0;
}

# wait for a file
sub wait_for_file {
    my ($sys,$file)=@_;
    while (!-f $file) {
        my $sleep++;
        if ($sleep%6000==0) { # should never hit, but will detect hangs
            $sys->log("Waited $sleep seconds for $file");
        }
        sleep(0.01);
    }
    return 0;
}

# wait for a file to be removed
# wait log period is much longer because a long wait is possible during install
sub wait_for_file_rm {
    my ($sys,$file)=@_;
    while (-f $file) {
        my $sleep++;
        if ($sleep%6000==0) { # should rarely hit
            $sys->log("Waited $sleep seconds for $file to be removed");
        }
        sleep(0.01);
    }
    return 0;
}

#get the application's current path
sub app_path {
    my $sys=shift;
    my ($cwd,$dir);
    $cwd=dirname($0);
    if ($cwd =~ m{^/}) {
        $dir=$cwd;
    } elsif ($ENV{PWD}) {
        if ($cwd eq '.') {
            $dir="$ENV{PWD}";
        } else {
            $dir="$ENV{PWD}/$cwd";
        }
    } else {
        $dir=abs_path($cwd);
    }
    return $dir;
}

sub mkdir {
    my ($sys,$dir) = @_;

    unless (-e $dir) {
        unless(File::Path::mkpath($dir)) {
            $sys->log("Could not create directory: $dir");
            #todo: return a code
            return;
        }
    }
    return 0;
}

sub rmdir {
    my ($sys,$dir) = @_;

    if (-e $dir) {
        unless(File::Path::rmtree($dir)) {
            $sys->log("Could not remove directory: $dir");
            #todo: return a code
            return;
        }
    }
    return 0;
}

sub log {
    my ($sys, $msg, $type) = @_;
    my($logmsg,$time);

    $logmsg = (ref($msg) =~ m/^Sys/) ? $msg->{msg} : $msg;
    $time = EDRu::current_time();
    if ($type =~ /info/i) {
        $type = "INFO: ";
    } elsif ($type =~ /warn/i) {
        $type = "WARNING: ";
    } elsif ($type =~ /err/i) {
        $type = "ERROR: ";
    } elsif ($type =~ /fatal/i) {
        $type = "FATAL: ";
    }
    $sys->appendfile("$time $type$logmsg", $sys->{logfile});
    return 0;
}

sub create_logfile {
    my $sys=shift;
    $sys->log("Module=$sys->{module}");
    $sys->log("ARGS=\"$sys->{arg}\"");
    return 0;
}

sub die {
    my ($sys, $msg) = @_;
    $sys->log($msg,'fatal');
    exit 1;
}

sub summary {
    my ($sys, $module, $all_result) = @_;
    my ($result, $rc, $changed, @status);
    $rc = $all_result->{rc};
    $rc = "no_return_code" unless (defined $rc);
    $changed = $all_result->{changed};
    $changed = "no_change_value" unless (defined $changed);
    if ($all_result->{status}) {
        @status = @{$all_result->{status}};
    } else {
        @status = ("no_status_value");
    }   
    $result = "$module: $rc $changed @status";
    $sys->appendfile($result, $sys->{summaryfile});
    $sys->appendfile($all_result->{results}, $sys->{output}) if (defined $all_result->{results});

}
package Sys::Linux;
@Sys::Linux::ISA = qw(Sys::Common);

sub init_cmds{
	my ( $sys ) = @_;
	my ($class,$distro,$version,$arch);
    # Define padvform specific commands
    $sys->{cmd}{awk}='/bin/awk';
    $sys->{cmd}{arch}='/bin/arch';
    $sys->{cmd}{cat}='/bin/cat';
    $sys->{cmd}{chgrp}='/bin/chgrp';
    $sys->{cmd}{chmod}='/bin/chmod';
    $sys->{cmd}{chown}='/bin/chown';
    $sys->{cmd}{cp}='/bin/cp -L';
    $sys->{cmd}{cpp}='/bin/cp -P';
    $sys->{cmd}{cut}='/bin/cut';
    $sys->{cmd}{date}='/bin/date';
    $sys->{cmd}{dd}='/bin/dd';
    $sys->{cmd}{dfk}='/bin/df -klP';
    $sys->{cmd}{diff}='/usr/bin/diff';
    $sys->{cmd}{dirname}='/usr/bin/dirname';
    $sys->{cmd}{du}='/usr/bin/du';
    $sys->{cmd}{echo}='/bin/echo';
    $sys->{cmd}{egrep}='/bin/egrep';
    $sys->{cmd}{ethtool}='/sbin/ethtool';
    $sys->{cmd}{ibstatus}='/usr/sbin/ibstatus';
    $sys->{cmd}{ibstat}='/usr/sbin/ibstat';
    $sys->{cmd}{find}='/usr/bin/find';
    $sys->{cmd}{grep}='/bin/grep';
    $sys->{cmd}{groups}='/usr/bin/groups';
    $sys->{cmd}{groupadd}='/usr/sbin/groupadd';
    $sys->{cmd}{groupdel}='/usr/sbin/groupdel';
    $sys->{cmd}{groupmod}='/usr/sbin/groupmod';
    $sys->{cmd}{gunzip}='/bin/gunzip';
    $sys->{cmd}{head}='/usr/bin/head';
    $sys->{cmd}{hostname}='/bin/hostname';
    $sys->{cmd}{id}='/usr/bin/id';
    $sys->{cmd}{ip}='/sbin/ip';
    $sys->{cmd}{ipcalc}='/bin/ipcalc';
    $sys->{cmd}{ifconfig}='/sbin/ifconfig';
    $sys->{cmd}{kill}='/bin/kill';
    $sys->{cmd}{ln}='/bin/ln';
    $sys->{cmd}{ls}='/bin/ls';
    $sys->{cmd}{lsmod}='/sbin/lsmod';
    $sys->{cmd}{make}='/usr/bin/make';
    $sys->{cmd}{mkdir}='/bin/mkdir';
    $sys->{cmd}{mkdirp}='/bin/mkdir -p';
    $sys->{cmd}{modinfo}='/sbin/lsmod';
    $sys->{cmd}{modunload}='/sbin/modprobe -r';
    $sys->{cmd}{modprobe}='/sbin/modprobe';
    $sys->{cmd}{mount}='/bin/mount';
    $sys->{cmd}{mv}='/bin/mv';
    $sys->{cmd}{netstat}='/bin/netstat';
    $sys->{cmd}{nohup}='/usr/bin/nohup';
    $sys->{cmd}{nslookup}='/usr/bin/nslookup';
    $sys->{cmd}{ntpdate}='/usr/sbin/ntpdate';
    $sys->{cmd}{openssl}='/usr/bin/openssl';
	$sys->{cmd}{ping}='/bin/ping';
    $sys->{cmd}{ping6}='/bin/ping6';
    $sys->{cmd}{ps}='/bin/ps -w';
    $sys->{cmd}{rcp}='/usr/bin/rcp';
    $sys->{cmd}{readlink}='/bin/readlink';
    $sys->{cmd}{rm}='/bin/rm';
    $sys->{cmd}{rmr}='/bin/rm -rf';
    $sys->{cmd}{rmdir}='/bin/rm -rf';
    $sys->{cmd}{rmmod}='/sbin/rmmod';
    $sys->{cmd}{rpm}='/bin/rpm';
    $sys->{cmd}{rsh}='/usr/bin/rsh';
    $sys->{cmd}{scp}='/usr/bin/scp';
    $sys->{cmd}{sed}='/bin/sed';
    $sys->{cmd}{sh}='/bin/sh';
    $sys->{cmd}{shutdown}='/sbin/shutdown -r now';
    $sys->{cmd}{sleep}='/bin/sleep';
    $sys->{cmd}{sort}='/bin/sort';
    $sys->{cmd}{ssh}='/usr/bin/ssh';
    $sys->{cmd}{sshkeygen}='/usr/bin/ssh-keygen';
    $sys->{cmd}{sshkeyscan}='/usr/bin/ssh-keyscan';
    $sys->{cmd}{strings}='/usr/bin/strings';
    $sys->{cmd}{su}='/bin/su';
    $sys->{cmd}{tail}='/usr/bin/tail';
    $sys->{cmd}{tar}='/bin/tar';
    $sys->{cmd}{tee}='/usr/bin/tee';
    $sys->{cmd}{touch}='/bin/touch';
    $sys->{cmd}{tput}='/usr/bin/tput';
    $sys->{cmd}{tr}='/usr/bin/tr';
    $sys->{cmd}{uname}='/bin/uname';
    $sys->{cmd}{uniq}='/usr/bin/uniq';
    $sys->{cmd}{useradd}='/usr/sbin/useradd';
    $sys->{cmd}{userdel}='/usr/sbin/userdel';
    $sys->{cmd}{usermod}='/usr/sbin/usermod';
    $sys->{cmd}{vxdmpadm}='/sbin/vxdmpadm';
    $sys->{cmd}{vxkeyless}='/opt/VRTSvlic/bin/vxkeyless';
    $sys->{cmd}{vxlicrep}='/sbin/vxlicrep';
    $sys->{cmd}{vxlicinst}='/sbin/vxlicinst';
    $sys->{cmd}{vxlicinstupgrade}='/sbin/vxlicinstupgrade';
    $sys->{cmd}{vxtune}='/sbin/vxtune';
    $sys->{cmd}{wc}='/usr/bin/wc';
    $sys->{cmd}{which}='/usr/bin/which';
    $sys->{cmd}{yes}='/usr/bin/yes';
    $sys->{cmd}{hostid}='/usr/bin/hostid';
    $sys->{cmd}{nm}='/usr/bin/nm';
    $sys->{cmd}{getconf}='/usr/bin/getconf';
    $sys->{cmd}{dmesg}='/bin/dmesg';
    $sys->{cmd}{dmidecode}='/usr/sbin/dmidecode';
    $sys->{cmd}{cksum}='/usr/bin/cksum';
    $sys->{cmd}{service}='/sbin/service';
	$sys->{cmd}{lspci}='/sbin/lspci -v';
    $sys->{cmd}{udevadm}='/sbin/udevadm';
    $sys->{cmd}{route}='/sbin/route';
    $sys->{cmd}{lshal}='/usr/bin/lshal';
    $sys->{cmd}{crontab}='/usr/bin/crontab';
    $sys->{cmd}{systemctl}='/usr/bin/systemctl';
    
    #Product commands
    $sys->{cmd}{stat}='/usr/bin/stat';
    $sys->{cmd}{mknod}='/bin/mknod';
    $sys->{cmd}{vxdctl}='/sbin/vxdctl';
    $sys->{cmd}{vxdisk}='/sbin/vxdisk';
    $sys->{cmd}{vxinfo}='/usr/sbin/vxinfo';
    $sys->{cmd}{vxconfigd}='/sbin/vxconfigd';
    $sys->{cmd}{vxddladm}='/usr/sbin/vxddladm';
    $sys->{cmd}{vxrelocd}='/usr/lib/vxvm/bin/vxrelocd';
    $sys->{cmd}{vxattachd}='/usr/lib/vxvm/bin/vxattachd';
    $sys->{cmd}{vxdg}='/usr/sbin/vxdg';
    $sys->{cmd}{vxprint}='/usr/sbin/vxprint';
    $sys->{cmd}{vxscriptlog}='/usr/sbin/vxscriptlog';
    $sys->{cmd}{vxdmpinit}='/etc/vx/bin/vxdmpinit';
    $sys->{cmd}{vxedit}='/usr/sbin/vxedit';
    $sys->{cmd}{vxvol}='/usr/sbin/vxvol';
    $sys->{cmd}{vxassist}='/usr/sbin/vxassist';
    $sys->{cmd}{vxrecover}='/usr/sbin/vxrecover';
    $sys->{cmd}{vxrvg}='/usr/sbin/vxrvg';
    $sys->{cmd}{vxrlink}='/usr/sbin/vxrlink';
    $sys->{cmd}{vradmin}='/usr/sbin/vradmin';
    $sys->{cmd}{vxrecover}='/usr/sbin/vxrecover';
    $sys->{cmd}{vras_script}='/etc/init.d/vras-vradmind.sh';
    $sys->{cmd}{vxclustadm}='/opt/VRTS/bin/vxclustadm';
    $sys->{cmd}{vxcached}='/usr/lib/vxvm/bin/vxcached';
    $sys->{cmd}{vxconfigbackupd}='/etc/vx/bin/vxconfigbackupd';
    $sys->{cmd}{vxgetrootdisk}='/etc/vx/bin/vxgetrootdisk';
    $sys->{cmd}{vxdisksetup}='/opt/VRTS/bin/vxdisksetup';
    $sys->{cmd}{sfcache}='/usr/sbin/sfcache';
    $sys->{cmd}{vxvvrstatd}='/sbin/vxvvrstatd';
	$sys->{cmd}{fstyp}='/opt/VRTS/bin/fstyp';
	$sys->{cmd}{cfscluster}="/opt/VRTS/bin/cfscluster";
    $sys->{cmd}{cfsdgadm}="/opt/VRTS/bin/cfsdgadm";
    $sys->{cmd}{cfsmntadm}="/opt/VRTS/bin/cfsmntadm";
    $sys->{cmd}{cfsmount}="/opt/VRTS/bin/cfsmount";
    $sys->{cmd}{cfsumount}="/opt/VRTS/bin/cfsumount";
    $sys->{cmd}{fsclustadm}="/opt/VRTS/bin/fsclustadm";
    $sys->{cmd}{vcsmmconfig}='/sbin/vcsmmconfig';
    $sys->{cmd}{odmclusteradm}='/opt/VRTSodm/bin/odmclustadm';
    $sys->{cmd}{cfsshare}="/opt/VRTS/bin/cfsshare";
    $sys->{cmd}{vxprint}='/usr/sbin/vxprint';
	$sys->{cmd}{mkfs}='/sbin/mkfs';
    $sys->{cmd}{format}='/usr/sbin/format';
    $sys->{cmd}{psrinfo} = '/usr/sbin/psrinfo';
    $sys->{cmd}{prctl} = '/usr/bin/prctl';
    $sys->{cmd}{sysdef} = '/usr/sbin/sysdef';
	$sys->{cmd}{lltstat} = '/sbin/lltstat';
    $sys->{cmd}{lltconfig} = '/sbin/lltconfig';
    $sys->{cmd}{gabconfig} = '/sbin/gabconfig';
    $sys->{cmd}{amfconfig} = '/opt/VRTSamf/bin/amfconfig';
    $sys->{cmd}{vxfenconfig} = '/sbin/vxfenconfig';
    $sys->{cmd}{vxfenadm} = '/sbin/vxfenadm';
    $sys->{cmd}{hastop} = '/opt/VRTSvcs/bin/hastop';
    $sys->{cmd}{hasys} = '/opt/VRTSvcs/bin/hasys';
    $sys->{cmd}{hagrp} = '/opt/VRTSvcs/bin/hagrp';
    $sys->{cmd}{hares} = '/opt/VRTSvcs/bin/hares';
    $sys->{cmd}{hastatus} = '/opt/VRTSvcs/bin/hastatus';
	$sys->{cmd}{haconf} = '/opt/VRTSvcs/bin/haconf';
	$sys->{cmd}{haclus} = '/opt/VRTSvcs/bin/haclus';
	$sys->{cmd}{vvrport}='/usr/sbin/vrport';
	$sys->{cmd}{vvrstart}='/usr/sbin/vxstart_vvr';
	$sys->{cmd}{vvrstop}='/usr/sbin/vxstart_vvr stop';
	$sys->{cmd}{vras_script}='/etc/init.d/vras-vradmind.sh';
	$sys->{cmd}{vxrootadm}='/etc/vx/bin/vxrootadm';
	$sys->{cmd}{vxvset}='/usr/sbin/vxvset';
	$sys->{cmd}{vxfen}='/etc/init.d/vxfen';
	$sys->{cmd}{vcs}='/etc/init.d/vcs';
	$sys->{cmd}{yum}='/usr/bin/yum';
	$sys->{cmd}{env}='/bin/env';
	$sys->{cmd}{kstat}='/usr/bin/kstat';

	if ( $sys->{class} =~ /^Sys\::(RHEL|SLES|Debian|CentOS|OL|ESX)(\d+)(.*)/m ){
		$distro=$1;
		$version=$2;
		$arch=$3;
    	$arch='64 bit' if ($arch eq 'x8664');
									
		if ($distro eq 'SLES') {
	        $sys->{cmd}{basename}='/usr/bin/basename';
	        $sys->{cmd}{cut}='/usr/bin/cut';
	        $sys->{cmd}{gunzip}='/usr/bin/gunzip';
	        # On SLES11 and above, ethtool path is '/sbin/ethtool'.
            $sys->{cmd}{ethtool}='/usr/sbin/ethtool' if ($version < 11); 
        }
	}
	return;
}

package Sys::RHEL6x8664;
@Sys::RHEL6x8664::ISA = qw(Sys::Linux);

package Sys::OL6x8664;
@Sys::OL6x8664::ISA = qw(Sys::Linux);

package Sys::RHEL7x8664;
@Sys::RHEL7x8664::ISA = qw(Sys::Linux);

package Sys::OL7x8664;
@Sys::OL7x8664::ISA = qw(Sys::Linux);

1;
